import os
from genericpath import getmtime
from os.path import dirname, join, exists, isfile, splitext, basename, isdir, relpath

from datetime import datetime
from collections import namedtuple
from nose import SkipTest

from ngs_utils.file_utils import add_suffix
from ngs_utils.testing import BaseTestCase, swap_output, check_call
from tests import BaseTargQC, info


class AnnotateBedTests(BaseTestCase):
    script = 'annotate_bed.py'

    data_dir = join(dirname(__file__), BaseTestCase.data_dir, 'bed')
    results_dir = join(dirname(__file__), BaseTestCase.results_dir, 'annotate_bed')
    gold_standard_dir = join(join(dirname(__file__), BaseTestCase.gold_standard_dir), 'annotate_bed')

    def test_hg19(self):
        self._test('hg19', 'hg19')

    def test_hg38(self):
        self._test('hg38', 'hg38')

    def test_hg38_seq2c(self):
        self._test('hg38_seq2c', 'hg38', ['--seq2c'])

    def test_hg38_ext(self):
        self._test('hg38_ext', 'hg38', ['--extended'])

    def test_hg38_ext_features(self):
        self._test('hg38_ext_features', 'hg38', ['--extended', '--output-features'])

    def test_hg38_ext_features_highconf(self):
        self._test('hg38_ext_features_high_conf', 'hg38', ['--extended', '--output-features', '--high-confidence'])

    def test_hg38_short(self):
        self._test('hg38_short', 'hg38', ['--short'])

    def test_mm10(self):
        self._test('mm10', 'mm10', ['--extended'])

    def _test(self, name, genome, opts=None):
        os.chdir(self.results_dir)
        input_fname = genome + '.bed'
        output_fname = add_suffix(input_fname, 'anno')
        input_fpath = join(self.data_dir, input_fname)
        output_fpath = join(self.results_dir, output_fname)

        cmdl = [self.script,
                input_fpath,
                '-o', output_fpath] + \
               (opts or []) + \
               ['--debug'] + \
               ['-g', genome]

        swap_output(output_fpath)

        info('-' * 100)
        check_call(cmdl)
        info('-' * 100)
        info('')

        self._check_file(output_fpath)




import shutil
import os
from os import listdir
from os.path import relpath, join, exists, dirname, basename, abspath, splitext
from collections import OrderedDict, defaultdict

import targqc
import targqc.config as tc
from ngs_utils.file_utils import verify_dir, verify_file, adjust_path, symlink_plus, file_transaction, add_suffix
from ngs_utils.logger import info, err, debug
from ngs_utils.reporting.reporting import PerRegionSampleReport, BaseReport, Metric, ReportSection, MetricStorage, \
    FullReport
from targqc.general_report import get_header_metric_storage
from targqc.qualimap.runner import run_multisample_qualimap


def _make_targetcov_symlinks(samples):
    for sample in samples:
        new_link = join(
            dirname(dirname(sample.targetcov_detailed_txt)),
            basename(sample.targetcov_detailed_txt))
        if exists(new_link):
            os.unlink(new_link)
        symlink_plus(sample.targetcov_detailed_txt, new_link)
        info('TargetCov TXT symlink saved to ' + new_link)


def make_tarqc_html_report(output_dir, work_dir, samples, bed_fpath=None, tag_by_sample=None):
    # header_storage = get_header_metric_storage(tc.depth_thresholds,
    #                                            is_wgs=bed_fpath is not None,
    #                                            padding=tc.padding)

    jsons_by_sample = {s.name: s.targqc_json_fpath for s in samples if verify_file(s.targqc_json_fpath)}
    # htmls_by_sample = {s.name: s.targqc_html_fpath for s in samples if verify_file(s.targqc_html_fpath)}
    htmls_by_sample = dict()

    if not jsons_by_sample:
        return None, None, None

    targqc_full_report = FullReport.construct_from_sample_report_jsons(samples, output_dir, jsons_by_sample, htmls_by_sample)

    for sample_report in targqc_full_report.sample_reports:
        if tag_by_sample:
            sample_report.set_project_tag(tag_by_sample[sample_report.sample.name])
        if verify_file(sample_report.sample.qualimap_html_fpath):
            url = relpath(sample_report.sample.qualimap_html_fpath, output_dir)
            r = sample_report.find_record(sample_report.records, 'Qualimap')
            if r:
                r.url = url
            else:
                sample_report.add_record(metric_name='Qualimap', value='Qualimap', url=url, silent=True)

    if len(samples) > 1:
        run_multisample_qualimap(output_dir, work_dir, samples, targqc_full_report)

    fn = splitext(basename(samples[0].targqc_txt_fpath))[0]
    tsv_fpath = targqc_full_report.save_tsv(join(output_dir, fn + '.tsv'))
    html_fpath = targqc_full_report.save_html(join(output_dir, fn + '.html'), 'TargQC')

    return tsv_fpath, html_fpath


def combined_regional_reports(work_dir, output_dir, samples):
    if not any(verify_file(s.targqc_region_tsv, silent=True) for s in samples):
        return None, None

    tsv_region_rep_fpath = join(output_dir, basename(samples[0].targqc_region_tsv))
    debug('Combining regional reports, writing to ' + tsv_region_rep_fpath)
    with file_transaction(work_dir, tsv_region_rep_fpath) as tx_tsv:
        with open(tx_tsv, 'w') as tsv_out:
            # sample_i = 0
            # for s in samples:
            #     if s.targqc_region_txt and verify_file(s.targqc_region_txt):
            #         with open(s.targqc_region_txt) as txt_in:
            #             for l in txt_in:
            #                 if l.startswith('#'):
            #                     if not l.startswith('##') and sample_i == 0:
            #                         txt_out.write('#Sample' + ' '*(max(len('#Sample'), len(s.name)) - len('#Sample')) + ' ' + l.replace('#Chr', 'Chr '))
            #                 else:
            #                     txt_out.write(s.name + ' '*(max(len('#Sample'), len(s.name)) - len(s.name)) + ' ' + l)
            #         sample_i += 1
            sample_i = 0
            for s in samples:
                if s.targqc_region_tsv and verify_file(s.targqc_region_tsv):
                    with open(s.targqc_region_tsv) as tsv_in:
                        for l in tsv_in:
                            if l.startswith('#'):
                                if not l.startswith('##') and sample_i == 0:
                                    tsv_out.write('#Sample\t' + l[1:])
                            else:
                                tsv_out.write(s.name + '\t' + l)
                    sample_i += 1

    return tsv_region_rep_fpath


# def summarize_targqc(summary_threads, output_dir, work_dir, samples, bed_fpath=None, tag_by_sample=None):
    # best_for_regions_fpath = None
    # if any(verify_file(s.targqc_region_tsv, silent=True) for s in samples):
    #     best_for_regions_fpath = _save_best_details_for_each_gene(tc.depth_thresholds, samples, output_dir)
    # ''' 1. best_regions = get_best_regions()
    #     2. best_for_regions_fpath = save_per_region_report()
    #     3. calc median coverage across best regions
    #     4. flagged_regions_report_fpath = _generate_flagged_regions_report(
    #          output_dir, 'Best', average_coverage, genes, depth_threshs)
    # '''


def _prep_best_report(metric_storage, samples):
    report = PerRegionSampleReport(sample='Best', metric_storage=metric_storage)

    report.add_record('Sample', 'contains best values from all samples: ' + ', '.join([s.name for s in samples]))

    m = metric_storage.find_metric('Average sample depth')
    ave_sample_depth = max(BaseReport.find_record(s.report.records, m.name).value for s in samples)
    report.add_record('Average sample depth', ave_sample_depth)

    return report


def _get_targqc_metric_storage(metric_storages_by_report_type):
    class SectionId:
        def __init__(self, name, title):
            self.name = name
            self.title = title

        def __hash__(self):
            #return hash((self.name, self.title))
            return hash(self.name)  # use title from the first metric_storage

        def __eq__(self, other):
            #return (self.name, self.title) == (other.name, other.title)
            return self.name == other.name  # use title from the first metric_storage

    metrics_by_sections = OrderedDict()
    general_section_id = None
    general_section_metric_list = []

    for report_type, metric_storage in metric_storages_by_report_type:
        for section in metric_storage.sections:
            section_id = SectionId(section.name, section.title)
            if section_id not in metrics_by_sections.keys():
                metrics_by_sections[section_id] = []

            metrics_by_sections[section_id] += [metric
                for metric in metric_storage.get_metrics(sections=[section], skip_general_section=True)
                if metric == _get_targqc_metric(metric, dict(metric_storages_by_report_type)['targetcov'], report_type)]

        # specific behaviour for general section
        general_section_metric_list += [metric
            for metric in metric_storage.general_section.metrics
            if metric == _get_targqc_metric(metric, dict(metric_storages_by_report_type)['targetcov'], report_type)]
        if not general_section_id:
            general_section_id = SectionId(metric_storage.general_section.name, metric_storage.general_section.title)

    sections = []
    for section_id, metric_list in metrics_by_sections.items():
        sections.append(ReportSection(section_id.name, section_id.title, metric_list))

    return MetricStorage(
        general_section=ReportSection(
            general_section_id.name, general_section_id.title, general_section_metric_list),
        sections=sections)


def _correct_qualimap_genome_results(samples):
    """ fixing java.lang.Double.parseDouble error on entries like "6,082.49"
    """
    for s in samples:
        if verify_file(s.qualimap_genome_results_fpath):
            correction_is_needed = False
            with open(s.qualimap_genome_results_fpath, 'r') as f:
                content = f.readlines()
                metrics_started = False
                for line in content:
                    if ">> Reference" in line:
                        metrics_started = True
                    if metrics_started:
                        if line.find(',') != -1:
                            correction_is_needed = True
                            break
            if correction_is_needed:
                with open(s.qualimap_genome_results_fpath, 'w') as f:
                    metrics_started = False
                    for line in content:
                        if ">> Reference" in line:
                            metrics_started = True
                        if metrics_started:
                            if line.find(',') != -1:
                                line = line.replace(',', '')
                        f.write(line)


def _correct_qualimap_insert_size_histogram(work_dir, samples):
    """ replacing Qualimap insert size histogram with Picard one.
    """
    for s in samples:
        qualimap1_dirname = dirname(s.qualimap_ins_size_hist_fpath).replace('raw_data_qualimapReport', 'raw_data')
        qualimap2_dirname = dirname(s.qualimap_ins_size_hist_fpath)
        if exists(qualimap1_dirname):
            if not exists(qualimap2_dirname):
                shutil.move(qualimap1_dirname, qualimap2_dirname)
            else:
                shutil.rmtree(qualimap1_dirname)
        elif not exists(qualimap2_dirname):
            continue  # no data from both Qualimap v.1 and Qualimap v.2

        # if qualimap histogram exits and reuse_intermediate, skip
        if verify_file(s.qualimap_ins_size_hist_fpath, silent=True) and tc.reuse_intermediate:
            pass
        else:
            if verify_file(s.picard_ins_size_hist_txt_fpath):
                with open(s.picard_ins_size_hist_txt_fpath, 'r') as picard_f:
                    one_line_to_stop = False
                    for line in picard_f:
                        if one_line_to_stop:
                            break
                        if line.startswith('## HISTOGRAM'):
                            one_line_to_stop = True

                    with file_transaction(work_dir, s.qualimap_ins_size_hist_fpath) as tx:
                        with open(tx, 'w') as qualimap_f:
                            for line in picard_f:
                                qualimap_f.write(line)


def select_best(values, fn=max):
    vs = [v for v in values if v is not None]
    return fn(vs) if len(vs) > 0 else None


def get_int_val(v):
    v = _get_num(v)
    return int(v) if v else None

def get_float_val(v):
    v = _get_num(v)
    return float(v) if v else None

def _get_num(v):
    v = get_val(v)
    return ''.join(c for c in v if c.isdigit() or c == '.') if v else None

def get_val(v):
    return v.strip() if v.strip() not in ['.', '-', ''] else None


def _save_best_details_for_each_gene(depth_threshs, samples, output_dir):
    metric_storage = get_detailed_metric_storage(depth_threshs)

    report = PerRegionSampleReport(sample='Best', metric_storage=metric_storage)
    report.add_record('Sample', 'contains best values from all samples: ' + ', '.join([s.name for s in samples]))

    total_regions = 0
    fpaths = [s.targqc_region_tsv for s in samples if verify_file(s.targqc_region_tsv)]
    if not fpaths:
        err('No targetcov detailed per-gene report was generated; skipping.')
        return None

    open_tsv_files = [open(fpath) for fpath in fpaths]

    first_col = 0
    while True:
        lines_for_each_sample = [next(f, None) for f in open_tsv_files]
        if not all(lines_for_each_sample):
            break
        l = lines_for_each_sample[0]
        if l.startswith('##'):
            continue
        elif l.startswith('#'):
            if l.startswith('#Sample'):
                first_col = 1
            break

    while True:
        lines_for_each_sample = [next(f, None) for f in open_tsv_files]
        if not all(lines_for_each_sample):
            break

        if all([not l.startswith('#') and ('Whole-Gene' in l or 'Gene-Exon' in l) for l in lines_for_each_sample]):
            shared_fields = lines_for_each_sample[0].split('\t')[first_col:first_col+9]
            reg = report.add_row()
            reg.add_record('Chr', get_val(shared_fields[0]))
            reg.add_record('Start', get_int_val(shared_fields[1]))
            reg.add_record('End', get_int_val(shared_fields[2]))
            reg.add_record('Size', get_int_val(shared_fields[3]))
            reg.add_record('Gene', get_val(shared_fields[4]))
            reg.add_record('Strand', get_val(shared_fields[5]))
            reg.add_record('Feature', get_val(shared_fields[6]))
            reg.add_record('Biotype', get_val(shared_fields[7]))
            reg.add_record('Transcript', get_val(shared_fields[8]))

            min_depths, ave_depths, stddevs, withins = ([], [], [], [])
            percents_by_threshs = {t: [] for t in depth_threshs}

            for l in lines_for_each_sample:
                fs = l.split('\t')

                min_depths.append(get_int_val(fs[first_col+9]))
                ave_depths.append(get_float_val(fs[first_col+10]))
                stddevs.append(get_float_val(fs[first_col+11]))
                withins.append(get_float_val(fs[first_col+12]))
                for t, f in zip(depth_threshs, fs[first_col+13:]):
                    percents_by_threshs[t].append(get_float_val(f))

            # counting bests
            reg.add_record('Min depth', select_best(min_depths))
            reg.add_record('Ave depth', select_best(ave_depths))
            reg.add_record('Std dev', select_best(stddevs, max))
            reg.add_record('W/n 20% of median depth', select_best(withins))
            for t in depth_threshs:
                reg.add_record('{}x'.format(t), select_best(percents_by_threshs[t]))

            total_regions += 1

    for f in open_tsv_files:
        f.close()

    gene_report_basename = add_suffix(samples[0].targqc_region_tsv, 'best')
    txt_rep_fpath = report.save_txt(join(output_dir, gene_report_basename + '.txt'))
    tsv_rep_fpath = report.save_tsv(join(output_dir, gene_report_basename + '.tsv'))
    info('')
    info('Best values for the regions (total ' + str(total_regions) + ') saved into:')
    info('  ' + txt_rep_fpath)

    return txt_rep_fpath







import sys
import os
from os.path import dirname, join, abspath, isfile, pardir
from pybedtools import BedTool

from ngs_utils.bed_utils import bedtools_version
from ngs_utils.file_utils import which, open_gzipsafe, verify_file
from ngs_utils.logger import debug, critical


SUPPORTED_GENOMES = ['hg19', 'hg19-noalt', 'hg38', 'hg38-noalt']

class BedCols:
    CHROM, \
    START, \
    END, \
    GENE, \
    EXON, \
    STRAND, \
    FEATURE, \
    BIOTYPE, \
    ENSEMBL_ID, \
    TSL, \
    HUGO, \
    TX_OVERLAP_PERCENTAGE, \
    EXON_OVERLAPS_PERCENTAGE, \
    CDS_OVERLAPS_PERCENTAGE \
        = cols = range(14)

    names = {
        CHROM: '#Chrom',
        START: 'Start',
        END: 'End',
        GENE: 'Gene',
        EXON: 'Exon',
        STRAND: 'Strand',
        FEATURE: 'Feature',
        BIOTYPE: 'Biotype',
        ENSEMBL_ID: 'Ensembl_ID',
        TSL: 'TSL',
        HUGO: 'HUGO',
        # TX_OVERLAP_BASES: 'Tx_overlap_bp',
        TX_OVERLAP_PERCENTAGE: 'Tx_overlap_%',
        # EXON_OVERLAPS_BASES: 'Exon_overlaps_bp',
        EXON_OVERLAPS_PERCENTAGE: 'Exon_overlaps_%',
        CDS_OVERLAPS_PERCENTAGE: 'CDS_overlaps_%',
    }

def check_genome(genome):
    if genome not in SUPPORTED_GENOMES:
        sys.stdout.write('Genome ' + genome + ' is not supported. Supported genomes: ' + ', '.join(SUPPORTED_GENOMES) + '\n')
        sys.exit(1)

#################
### INTERFACE ###
#################
def get_all_features(genome, high_confidence=False, features=None, gene_names=None, only_canonical=False):
    _canon_filt = get_only_canonical_filter(genome) if only_canonical else None
    
    bed = _get_ensembl_file('ensembl.bed', genome)
    def _filter(x):
        if high_confidence:
            if x[BedCols.HUGO] in ['', '.', None]:
                return False
        if features:
            if x[BedCols.FEATURE] not in features:
                return False
        if gene_names:
            if x[BedCols.GENE] not in gene_names:
                return False
        if _canon_filt:
            if not _canon_filt(x):
                return False
        return True
    bed = bed.filter(_filter)
    return bed

def get_merged_cds(genome):
    """
    Returns all CDS merged, used:
    - for TargQC general reports CDS coverage statistics for WGS
    - for Seq2C CNV calling when no capture BED available
    """
    bed = get_all_features(genome)
    return bed\
        .filter(lambda r: r.fields[BedCols.FEATURE] in ['CDS', 'stop_codon'])\
        .filter(high_confidence_filter)\
        .merge()

###############
### ENSEMBL ###
###############
def ensembl_gtf_fpath(genome):
    return _get_ensembl_file(join('gtf', 'ref-transcripts.gtf'), genome.split('-')[0])  # no -alt

def biomart_fpath(genome='hg38'):
    """ bm_fpath downloaded from http://www.ensembl.org/biomart

        ---------------------------------------------------------
        hg38:
        - go to http://grch37.ensembl.org/biomart
        - select "Ensembl Gene"
        - select "Human genes (GRCh37.p13)"
        - click "Attributes":
          - GENE:
            - Transcript ID
            - Transcript Support Level (TSL)
            - Associated Gene Name
            - % GC content
            - Gene type
            - Transcript type
          - EXTERNAL:
            - HGNC symbol
            - RefSeq mRNA
            - RefSeq ncRNA

        link: http://www.ensembl.org/biomart/martview/1a58c2026b9aafb613b19630264d4a54?VIRTUALSCHEMANAME=default
        &ATTRIBUTES=hsapiens_gene_ensembl.default.feature_page.ensembl_transcript_id|hsapiens_gene_ensembl.default
        .feature_page.gene_biotype|hsapiens_gene_ensembl.default.feature_page.transcript_biotype
        |hsapiens_gene_ensembl.default.feature_page.transcript_tsl|hsapiens_gene_ensembl.default.feature_page
        .percentage_gc_content|hsapiens_gene_ensembl.default.feature_page.hgnc_symbol|hsapiens_gene_ensembl.default
        .feature_page.refseq_mrna|hsapiens_gene_ensembl.default.feature_page.refseq_ncrna|hsapiens_gene_ensembl
        .default.feature_page.external_gene_name&FILTERS=&VISIBLEPANEL=attributepanel
        
        ---------------------------------------------------------
        hg19:
        - go to http://www.ensembl.org/biomart
        - select "Ensembl Genes 87"
        - select "Human genes (GRCh38.p7)"
        - click "Attributes":
          - GENE:
            - Transcript ID
            - Associated Gene Name
            - % GC content
            - Gene type
            - Transcript type
          - EXTERNAL:
            - HGNC symbol
            - RefSeq mRNA
            - RefSeq ncRNA

        link: http://grch37.ensembl.org/biomart/martview/2b820216c72bef00db4384dfc0e874fd?VIRTUALSCHEMANAME=default
        &ATTRIBUTES=hsapiens_gene_ensembl.default.feature_page.ensembl_transcript_id|hsapiens_gene_ensembl.default
        .feature_page.external_gene_name|hsapiens_gene_ensembl.default.feature_page.percentage_gc_content
        |hsapiens_gene_ensembl.default.feature_page.gene_biotype|hsapiens_gene_ensembl.default.feature_page
        .transcript_biotype|hsapiens_gene_ensembl.default.feature_page.refseq_mrna|hsapiens_gene_ensembl.default
        .feature_page.refseq_ncrna|hsapiens_gene_ensembl.default.feature_page.hgnc_symbol&FILTERS=&VISIBLEPANEL
        =attributepanel
    """
    return _get_ensembl_file('mart_export.txt', genome)

def _get_ensembl_file(fname, genome=None):
    if genome:
        return _get(join(genome.split('-')[0], fname), genome)
    else:
        return _get(join(fname))


###################
### TRANSCRIPTS ###
###################
def get_canonical_transcripts_ids(genome):
    short_genome = genome.split('-')[0]
    if short_genome.startswith('GRCh37'):
        short_genome = 'hg19'
    if short_genome.startswith('GRCh38'):
        short_genome = 'hg38'
    check_genome(short_genome)

    canon_fpath = _get(join('{genome}', 'canon_transcripts_{genome}_ensembl.txt'), genome)
    replacement_fpath = _get('canon_cancer_replacement.txt')

    canon_fpath = verify_file(canon_fpath, is_critical=True, description='Canonical transcripts path')
    replacement_fpath = verify_file(replacement_fpath, is_critical=True, description='Canonical cancer transcripts replacement path')

    with open(canon_fpath) as f:
        canon_tx_by_gname = dict(l.strip('\n').split('\t') for l in f)
    with open(replacement_fpath) as f:
        for gname, tx_id in (l.strip('\n').split('\t') for l in f):
            canon_tx_by_gname[gname] = tx_id

    return canon_tx_by_gname


def _get(relative_path, genome=None):
    """
    :param relative_path: relative path of the file inside the repository
    :param genome: genome name. Can contain chromosome name after comma, like hg19-chr20,
                   in case of BED, the returning BedTool will be with added filter.
    :return: BedTools object if it's a BED file, or filepath
    """
    chrom = None
    if genome:
        if '-chr' in genome:
            genome, chrom = genome.split('-')
        check_genome(genome)
        relative_path = relative_path.format(genome=genome)

    path = abspath(join(dirname(__file__), relative_path))
    if not isfile(path) and isfile(path + '.gz'):
        path += '.gz'

    if path.endswith('.bed') or path.endswith('.bed.gz'):
        if path.endswith('.bed.gz'):
            bedtools = which('bedtools')
            if not bedtools:
                critical('bedtools not found in PATH: ' + str(os.environ['PATH']))
            bedtools_v = bedtools_version(bedtools)
            if bedtools_v > 25:
                debug('BED is compressed, creating BedTool')
                bed = BedTool(path)
            else:
                debug('BedTools version is < ' + str(bedtools_v) + ', extracting BED file')
                bed = BedTool(open_gzipsafe(path))
        else:
            debug('BED is uncompressed, creating BedTool')
            bed = BedTool(path)

        if chrom:
            bed = bed.filter(lambda r: r.chrom == chrom)
        return bed
    else:
        return path

def get_hgnc_gene_synonyms():
    return _get('HGNC_gene_synonyms.txt')

def high_confidence_filter(x):
    return x[BedCols.TSL] in ['1', '2', 'NA', '.', None] and x[BedCols.HUGO] not in ['', '.', None]

def get_only_canonical_filter(genome):
    canon_tx_by_gname = get_canonical_transcripts_ids(genome)
    return lambda x: x[BedCols.ENSEMBL_ID] == canon_tx_by_gname.get(x[BedCols.HUGO]) or \
                     x[BedCols.ENSEMBL_ID] == canon_tx_by_gname.get(x[BedCols.GENE])

def protein_coding_filter(x):
    return x[BedCols.BIOTYPE] == 'protein_coding'



import os
import sys
import shutil
from os.path import dirname, join, exists, isfile, splitext, basename, isdir, relpath, getsize, getmtime
from datetime import datetime
from collections import namedtuple

from ngs_utils.testing import BaseTestCase, info, check_call, swap_output


class BaseTargQC(BaseTestCase):
    script = 'targqc'

    data_dir = join(dirname(__file__), BaseTestCase.data_dir)
    results_dir = join(dirname(__file__), BaseTestCase.results_dir)
    gold_standard_dir = join(join(dirname(__file__), BaseTestCase.gold_standard_dir))

    syn3_url = 'http://quast.bioinf.spbau.ru/static/chr21.tar.gz'
    bwa_url = 'http://quast.bioinf.spbau.ru/static/bwa.tar.gz'

    syn3_dir = join(data_dir, 'chr21')
    bed3 = join(syn3_dir, 'NGv3.chr21.3col.bed')
    bed4 = join(syn3_dir, 'NGv3.chr21.4col.bed')
    Sample = namedtuple('Sample', 'name bam l_fastq r_fastq')
    samples = [
        Sample('syn3-tumor', 'syn3-tumor.bam', 'syn3-tumor_R1.fq.gz', 'syn3-tumor_R2.fq.gz'),
        Sample('syn3-normal', 'syn3-normal.bam', 'syn3-normal_R1.fq.gz', 'syn3-normal_R2.fq.gz'),
    ]

    bams = [join(syn3_dir, s.bam) for s in samples]
    fastqs = [join(syn3_dir, s.l_fastq) for s in samples] + [join(syn3_dir, s.r_fastq) for s in samples]
    bwa_dir = join(data_dir, 'bwa')
    bwa_path = join(bwa_dir, 'hg19-chr21.fa')

    def setUp(self):
        BaseTestCase.setUp(self)
        if not isdir(self.syn3_dir):
            info(self.syn3_dir + ' does not exist, downloading test data')
            cur_dir = os.getcwd()
            os.chdir(self.data_dir)
            check_call(['wget', self.syn3_url])
            check_call(['tar', '-xzvf', basename(self.syn3_url)])
            os.chdir(cur_dir)
        if not isdir(self.bwa_dir):
            info(self.bwa_dir + ' does not exist, downloading bwa reference data')
            cur_dir = os.getcwd()
            os.chdir(self.data_dir)
            check_call(['wget', self.bwa_url])
            check_call(['tar', '-xzvf', basename(self.bwa_url)])
            os.chdir(cur_dir)

    def _test(self, output_dirname=None, used_samples=samples, bams=None, fastq=None, bed=None,
              debug=False, reuse_intermediate=False, reuse_output_dir=False, reannotate=False,
              genome='hg19-chr21', bwa=None, threads=None, ipython=None, keep_work_dir=True):
        os.chdir(self.results_dir)
        cmdl = [self.script]
        output_dir = None
        if output_dirname:
            output_dir = join(self.results_dir, output_dirname)
            cmdl.extend(['-o', output_dir])
        if bams: cmdl.extend(bams)
        if fastq: cmdl.extend(fastq)
        if bed: cmdl.extend(['--bed', bed])
        if debug: cmdl.append('--debug')
        if reuse_intermediate:
            reuse_output_dir = True
            cmdl.append('--reuse')
        if reannotate: cmdl.append('--reannotate')
        if genome: 
            cmdl.extend(['-g', genome])
        if bwa:
            cmdl.extend(['--bwa', bwa])
        if threads: cmdl.extend(['-t', str(threads)])
        if ipython: cmdl.extend('-s sge -q queue -r pename=smp -r --local'.split())
        if keep_work_dir: cmdl.append('--keep-work-dir')

        output_dir = output_dir or self._default_output_dir()
        if reuse_output_dir is False:
            swap_output(output_dir)

        info('-' * 100)
        check_call(cmdl)
        info('-' * 100)
        info('')

        self._check_results(output_dir, used_samples)

        if self.remove_work_dir_on_success and not reuse_intermediate and not reuse_output_dir:
            work_dir = join(output_dir, 'work')
            if not isdir(work_dir):
                info('Work dir for run ' + output_dirname + ' does not exist under ' + work_dir)
            else:
                shutil.rmtree(work_dir)
        info('')

    @staticmethod
    def _default_output_dir():
        return join(os.getcwd(), 'targqc')

    def _check_results(self, output_dir, used_samples):
        assert isdir(output_dir)
        self._check_file_throws(join(output_dir, 'regions.tsv'))
        self._check_file_throws(join(output_dir, 'summary.tsv'))
        self._check_file_throws(join(output_dir, 'summary.html'), ignore_matching_lines='report_date')
        for s in used_samples:
            s_dir = join(output_dir, s.name)
            assert isdir(s_dir)
            self._check_file_throws(join(s_dir, 'regions.tsv'))
            self._check_file_throws(join(s_dir, 'summary.txt'))
            self._check_file_throws(join(s_dir, 'summary.html'), ignore_matching_lines='report_date')
            self._check_file_throws(join(s_dir, 'summary.json'), ignore_matching_lines='work_dir')
        # TODO: check line numbers and some values isntead of diff?



import os
from os.path import join, splitext, dirname

import ngs_utils.reference_data as ref
from ngs_utils.Sample import BaseSample
from ngs_utils.file_utils import safe_mkdir, can_reuse
from ngs_utils.sambamba import index_bam
from ngs_utils.logger import info, critical, debug
from ngs_utils import logger

from targqc import config
from targqc.fastq import proc_fastq
from targqc.region_coverage import make_region_reports
from targqc.general_report import make_general_reports
from targqc.Target import Target
from targqc.summarize import make_tarqc_html_report, combined_regional_reports

targqc_repr              = 'TargQC'
targqc_name              = 'targqc'

qualimap_name                   = 'qualimap'
qualimap_report_fname           = 'qualimapReport.html'
qualimap_genome_results_fname   = 'genome_results.txt'
qualimap_raw_data_dirname       = 'raw_data_qualimapReport'

qualimap_ishist_fname           = 'insert_size_histogram.txt'
qualimap_covhist_fname          = 'coverage_histogram.txt'
qualimap_gchist_fname           = 'mapped_reads_gc-content_distribution.txt'

picard_name              = 'picard'
picard_ishist_fname      = 'picard_ishist.txt'

fastqc_name              = 'fastqc'
dedup_bam                = 'dedup'  # work/post_processing/dedup and -ready.dedup.bam

fastqc_repr              = 'FastQC'
fastqc_report_fname      = 'fastqc_report.html'


from targqc.general_report import get_mean_cov as get_mean_cov
from config import depth_thresholds


def get_version():
    from targqc import version
    return version.__version__


def get_description():
    from targqc import version
    description = 'TargQC, target coverage evaluation tool. Version ' + version.__version__
    if version.__git_revision__:
        description += ', revision ' + version.__git_revision__
    return description


def start_targqc(work_dir, output_dir, samples, target_bed_fpath, parallel_cfg, bwa_prefix,
                 fai_fpath=None,
                 genome=config.genome,
                 depth_threshs=config.depth_thresholds,
                 downsample_to=config.downsample_fraction,
                 padding=config.padding,
                 dedup=config.dedup,
                 num_pairs_by_sample=None,
                 reannotate=config.reannotate,
                 ):
    d = get_description()
    info('*'*len(d))
    info(d)
    info('*'*len(d))
    info()

    fai_fpath = fai_fpath or ref.get_fai(genome)
    target = Target(work_dir, output_dir, fai_fpath, padding=padding, bed_fpath=target_bed_fpath,
         reannotate=reannotate, genome=genome, is_debug=logger.is_debug)

    fastq_samples = [s for s in samples if not s.bam and s.l_fpath and s.r_fpath]

    from ngs_utils.parallel import parallel_view
    if fastq_samples:
        if not bwa_prefix:
            critical('--bwa-prefix is required when running from fastq')
        with parallel_view(len(fastq_samples), parallel_cfg, join(work_dir, 'sge_fastq')) as view:
            num_pairs_by_sample = proc_fastq(fastq_samples, view, work_dir, bwa_prefix,
                downsample_to, num_pairs_by_sample, dedup=dedup)

    info()
    for s in samples:
        if s.bam:
            info(s.name + ': using alignment ' + s.bam)

    with parallel_view(len(samples), parallel_cfg, join(work_dir, 'sge_bam')) as view:
        if all(can_reuse(s.bam + '.bai', s.bam) for s in samples):
            debug('BAM indexes exists')
        else:
            info('Indexing BAMs...')
            view.run(index_bam, [[s.bam] for s in samples])

        info('Making general reports...')
        make_general_reports(view, samples, target, genome, depth_threshs, padding, num_pairs_by_sample,
                             is_debug=logger.is_debug, reannotate=reannotate, fai_fpath=fai_fpath)

    info()
    info('*' * 70)
    tsv_fpath, html_fpath = make_tarqc_html_report(output_dir, work_dir, samples, bed_fpath=target_bed_fpath)
    info('TargQC summary saved in: ')
    info('  ' + html_fpath)
    info('  ' + tsv_fpath)

    info()
    with parallel_view(len(samples), parallel_cfg, join(work_dir, 'sge_bam')) as view:
        info('Making region-level reports...')
        make_region_reports(view, work_dir, samples, target, genome, depth_threshs)

    info()
    info('*' * 70)
    tsv_region_rep_fpath = combined_regional_reports(work_dir, output_dir, samples)

    info()
    info('*' * 70)
    info('TargQC summary saved in: ')
    info('  ' + html_fpath)
    info('  ' + tsv_fpath)
    info('Per-region coverage statistics saved into:')
    info('  ' + tsv_region_rep_fpath)

    return html_fpath

    # for general_report, per_gene_report, sample in zip(general_reports, per_gene_reports, samples):
    #     info('')
    #     info('*' * 70)
    #     if general_report.txt_fpath and verify_file(general_report.txt_fpath):
    #         info('Summary report: ' + general_report.txt_fpath)
    #     if per_gene_report:
    #         path = per_gene_report if isinstance(per_gene_report, basestring) else per_gene_report.txt_fpath
    #         if path and verify_file(path):
    #             info('Region-based report: ' + path)


class Sample(BaseSample):
    def __init__(self, name, dirpath, *args, **kwargs):
        BaseSample.__init__(self, name, dirpath, targqc_dirpath=dirpath, *args, **kwargs)
        self.targqc_norm_depth_vcf_txt   = None
        self.targqc_norm_depth_vcf_tsv   = None

        self.targqc_txt_fpath            = join(self.targqc_dirpath, 'summary.txt')
        self.targqc_html_fpath           = join(self.targqc_dirpath, 'summary.html')
        self.targqc_json_fpath           = join(self.targqc_dirpath, 'summary.json')
        self.targqc_region_txt           = join(self.targqc_dirpath, 'regions.txt')
        self.targqc_region_tsv           = join(self.targqc_dirpath, 'regions.tsv')

        self.qualimap_dirpath = join(self.targqc_dirpath, 'qualimap')
        self.qualimap_html_fpath            = join(self.qualimap_dirpath, qualimap_report_fname)
        self.qualimap_genome_results_fpath  = join(self.qualimap_dirpath, qualimap_report_fname)
        self.qualimap_raw_dirpath           = join(self.qualimap_dirpath, qualimap_raw_data_dirname)
        self.qualimap_ins_size_hist_fpath   = join(self.qualimap_raw_dirpath, qualimap_ishist_fname)
        self.qualimap_cov_hist_fpath        = join(self.qualimap_raw_dirpath, qualimap_covhist_fname)
        self.qualimap_gc_hist_fpath         = join(self.qualimap_raw_dirpath, qualimap_gchist_fname)

        self.picard_dirpath                 = join(self.targqc_dirpath, picard_name)
        self.picard_ins_size_hist_txt_fpath = join(self.picard_dirpath, picard_ishist_fname)
        self.picard_ins_size_hist_pdf_fpath = join(self.picard_dirpath, splitext(picard_ishist_fname)[0] + '.pdf')


class OldStyleSample(BaseSample):
    def __init__(self, name, dirpath, *args, **kwargs):
        BaseSample.__init__(self, name, dirpath, targqc_dirpath=dirpath, *args, **kwargs)
        self.targqc_norm_depth_vcf_txt   = None
        self.targqc_norm_depth_vcf_tsv   = None

        self.targqc_txt_fpath         = join(self.targqc_dirpath, name + '.targetSeq.txt')
        self.targqc_html_fpath        = join(self.targqc_dirpath, name + '.targetSeq.html')
        self.targqc_json_fpath        = join(self.targqc_dirpath, name + '.targetSeq.json')
        self.targqc_detailed_txt      = join(self.targqc_dirpath, name + '.targetSeq.details.gene.txt')
        self.targqc_detailed_tsv      = join(self.targqc_dirpath, name + '.targetSeq.details.gene.tsv')

        self.qualimap_dirpath = join(self.targqc_dirpath, 'qualimap')
        self.qualimap_html_fpath            = join(self.qualimap_dirpath, qualimap_report_fname)
        self.qualimap_genome_results_fpath  = join(self.qualimap_dirpath, qualimap_report_fname)
        self.qualimap_raw_dirpath           = join(self.qualimap_dirpath, qualimap_raw_data_dirname)
        self.qualimap_ins_size_hist_fpath   = join(self.qualimap_raw_dirpath, qualimap_ishist_fname)
        self.qualimap_cov_hist_fpath        = join(self.qualimap_raw_dirpath, qualimap_covhist_fname)
        self.qualimap_gc_hist_fpath         = join(self.qualimap_raw_dirpath, qualimap_gchist_fname)

        self.picard_dirpath                 = join(self.targqc_dirpath, picard_name)
        self.picard_ins_size_hist_txt_fpath = join(self.picard_dirpath, picard_ishist_fname)
        self.picard_ins_size_hist_pdf_fpath = join(self.picard_dirpath, splitext(picard_ishist_fname)[0] + '.pdf')



import os
import subprocess
from distutils.version import LooseVersion
from os.path import getsize, dirname, join, abspath, relpath, isdir, exists, splitext, basename, isfile
from os import listdir
import shutil

from ngs_utils.call_process import run
from ngs_utils.file_utils import safe_mkdir, verify_file, verify_dir, file_transaction, file_exists, intermediate_fname, \
    can_reuse
from ngs_utils.logger import info, warn, err, critical, debug
from ngs_utils.reporting.reporting import write_tsv_rows

import targqc.config as cfg
import targqc


def get_qualimap_max_mem(bam):
    mem_m = getsize(bam) / 3 / 1024 / 1024
    mem_m = min(max(mem_m, 1200), 16000)
    return mem_m


def find_executable():
    root_dirpath = abspath(join(dirname(__file__)))
    for dname in listdir(abspath(join(dirname(__file__)))):
        if isdir(join(root_dirpath, dname)) and dname.startswith('qualimap') and file_exists(join(root_dirpath, dname, 'qualimap')):
            return join(root_dirpath, dname, 'qualimap')
    critical('Error: could not find Qualimap executable')


def run_qualimap(work_dir, output_dir, output_fpaths, bam_fpath, genome, bed_fpath=None, threads=1):
    info('Analysing ' + bam_fpath)

    safe_mkdir(dirname(output_dir))
    safe_mkdir(output_dir)

    mem_cmdl = ''
    mem_m = get_qualimap_max_mem(bam_fpath)
    mem = str(int(mem_m)) + 'M'
    mem_cmdl = '--java-mem-size=' + mem

    cmdline = (find_executable() + ' bamqc --skip-duplicated -nt {threads} {mem_cmdl} -nr 5000 '
        '-bam {bam_fpath} -outdir {output_dir}')

    if genome.startswith('hg') or genome.startswith('GRCh'):
        cmdline += ' -gd HUMAN'
    if genome.startswith('mm'):
        cmdline += ' -gd MOUSE'

    if bed_fpath:
        cmdline += ' -gff {bed_fpath}'
        debug('Using amplicons/capture panel ' + bed_fpath)

    cmdline = cmdline.format(**locals())
    if not all(can_reuse(fp, [bam_fpath, bed_fpath] if bed_fpath else [bam_fpath]) for fp in output_fpaths):
        for fp in output_fpaths:
            if isfile(fp):
                os.remove(fp)
        run(cmdline, env_vars=dict(DISPLAY=None))
    if not all(verify_file(fp, cmp_f=[bam_fpath, bed_fpath] if bed_fpath else [bam_fpath]) for fp in output_fpaths):
        critical('Some of the QualiMap results were not generated')

    return output_dir


def fix_bed_for_qualimap(bed_fpath, qualimap_bed_fpath):
    with open(bed_fpath) as inn, open(qualimap_bed_fpath, 'w') as out:
        for l in inn:
            fields = l.strip().split('\t')

            if len(fields) < 3:
                continue
            try:
                int(fields[1]), int(fields[2])
            except ValueError:
                continue

            if len(fields) < 4:
                fields.append('.')

            if len(fields) < 5:
                fields.append('0')

            if len(fields) < 6:
                fields.append('+')

            out.write('\t'.join(fields) + '\n')


def run_multisample_qualimap(output_dir, work_dir, samples, targqc_full_report):
    """ 1. Generates Qualimap2 plots and put into plots_dirpath
        2. Adds records to targqc_full_report.plots
    """
    plots_dirpath = join(output_dir, 'plots')
    individual_report_fpaths = [s.qualimap_html_fpath for s in samples]
    if isdir(plots_dirpath) and not any(
            not can_reuse(join(plots_dirpath, f), individual_report_fpaths)
            for f in listdir(plots_dirpath) if not f.startswith('.')):
        debug('Qualimap miltisample plots exist - ' + plots_dirpath + ', reusing...')
    else:
        # Qualimap2 run for multi-sample plots
        if len([s.qualimap_html_fpath for s in samples if s.qualimap_html_fpath]) > 0:
            if find_executable() is not None and get_qualimap_type(find_executable()) == 'full':
                qualimap_output_dir = join(work_dir, 'qualimap_multi_bamqc')

                _correct_qualimap_genome_results(samples)
                _correct_qualimap_insert_size_histogram(samples)

                safe_mkdir(qualimap_output_dir)
                rows = []
                for sample in samples:
                    if sample.qualimap_html_fpath:
                        rows += [[sample.name, sample.qualimap_html_fpath]]

                data_fpath = write_tsv_rows(([], rows), join(qualimap_output_dir, 'qualimap_results_by_sample.tsv'))
                qualimap_plots_dirpath = join(qualimap_output_dir, 'images_multisampleBamQcReport')
                cmdline = find_executable() + ' multi-bamqc --data {data_fpath} -outdir {qualimap_output_dir}'.format(**locals())
                run(cmdline, env_vars=dict(DISPLAY=None),
                    checks=[lambda _1, _2: verify_dir(qualimap_output_dir)], reuse=cfg.reuse_intermediate)

                if not verify_dir(qualimap_plots_dirpath):
                    warn('Warning: Qualimap for multi-sample analysis failed to finish. TargQC will not contain plots.')
                    return None
                else:
                    if exists(plots_dirpath):
                        shutil.rmtree(plots_dirpath)
                    shutil.move(qualimap_plots_dirpath, plots_dirpath)
            else:
                warn('Warning: Qualimap for multi-sample analysis was not found. TargQC will not contain plots.')
                return None

    targqc_full_report.plots = []
    for plot_fpath in listdir(plots_dirpath):
        plot_fpath = join(plots_dirpath, plot_fpath)
        if verify_file(plot_fpath) and plot_fpath.endswith('.png'):
            targqc_full_report.plots.append(relpath(plot_fpath, output_dir))


def get_qualimap_type(tool_cmdline):
    """Qualimap supports multi-bamqc functionality only starting from v.2.0
    """
    if LooseVersion(_get_qualimap_version(tool_cmdline)) >= LooseVersion("2.0"):
        return "full"
    else:
        return "limited"


def _get_qualimap_version(tool_cmdline):
    cmdline = tool_cmdline + ' -version'  # actually, Qualimap doesn't have -version option

    version = None
    with subprocess.Popen(cmdline,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.STDOUT,
                          shell=True).stdout as stdout:
        out = stdout.read().strip()
        flag = "QualiMap v."
        if out.startswith(flag) >= 0:
            version = out.split(flag)[-1].strip()
    if not version:
        info('WARNING: could not determine Qualimap version, using 1.0')
        return '1.0'
    if version.split('.') > 2:  # only major version
        version = '.'.join(version.split('.')[:2])
    return version


def _correct_qualimap_genome_results(samples):
    """ fixing java.lang.Double.parseDouble error on entries like "6,082.49"
    """
    for s in samples:
        if verify_file(s.qualimap_genome_results_fpath):
            correction_is_needed = False
            with open(s.qualimap_genome_results_fpath, 'r') as f:
                content = f.readlines()
                metrics_started = False
                for line in content:
                    if ">> Reference" in line:
                        metrics_started = True
                    if metrics_started:
                        if line.find(',') != -1:
                            correction_is_needed = True
                            break
            if correction_is_needed:
                with open(s.qualimap_genome_results_fpath, 'w') as f:
                    metrics_started = False
                    for line in content:
                        if ">> Reference" in line:
                            metrics_started = True
                        if metrics_started:
                            if line.find(',') != -1:
                                line = line.replace(',', '')
                        f.write(line)


def _correct_qualimap_insert_size_histogram(samples):
    """ replacing Qualimap insert size histogram with Picard one.
    """
    for s in samples:
        qualimap1_dirname = dirname(s.qualimap_ins_size_hist_fpath).replace('raw_data_qualimapReport', 'raw_data')
        qualimap2_dirname = dirname(s.qualimap_ins_size_hist_fpath)
        if exists(qualimap1_dirname):
            if not exists(qualimap2_dirname):
                shutil.move(qualimap1_dirname, qualimap2_dirname)
            else:
                shutil.rmtree(qualimap1_dirname)
        elif not exists(qualimap2_dirname):
            continue  # no data from both Qualimap v.1 and Qualimap v.2

        # if qualimap histogram exits and reuse_intermediate, skip
        if verify_file(s.qualimap_ins_size_hist_fpath, silent=True) and cfg.reuse_intermediate:
            pass
        else:
            if verify_file(s.picard_ins_size_hist_txt_fpath):
                with open(s.picard_ins_size_hist_txt_fpath, 'r') as picard_f:
                    one_line_to_stop = False
                    for line in picard_f:
                        if one_line_to_stop:
                            break
                        if line.startswith('## HISTOGRAM'):
                            one_line_to_stop = True

                    with file_transaction(None, s.qualimap_ins_size_hist_fpath) as tx:
                        with open(tx, 'w') as qualimap_f:
                            for line in picard_f:
                                qualimap_f.write(line)



from os.path import abspath, dirname, pardir, join

code_base_path = abspath(join(dirname(abspath(__file__)), pardir))

# Defaults (can be altered here or via command line arguments):
padding = 200
depth_thresholds = [1, 5, 10, 20, 50, 100, 250, 500, 1000, 5000, 10000, 50000]
downsample_fraction = 0.05
downsample_pairs_num = 5e5
genome = 'hg19'
dedup = True

reuse_intermediate = False
is_debug = False
threads = 1
reannotate = False  # reannotate BED even if the number of columns is 4 or higher


from os.path import dirname, join, exists, isfile, splitext, basename, isdir, relpath

from nose import SkipTest

from tests import BaseTargQC, info


class UnitTests(BaseTargQC):
    def test_01_simple(self):
        self._test('simple', [self.samples[0]], bams=[self.bams[0]], bed=self.bed4)

    def test_02_bed3(self):
        self._test('bed3', bams=self.bams, bed=self.bed3)

    def test_03_bed4(self):
        self._test('bed4', bams=self.bams, bed=self.bed4)

    def test_04_bed4_reannotate(self):
        self._test('bed4_reannotate', bams=self.bams, bed=self.bed4, reannotate=True)

    def test_05_wgs(self):
        self._test('wgs', bams=self.bams)

    def test_06_threads(self):
        self._test('threads', bams=self.bams, bed=self.bed4, threads='2')

    def test_07_ipython(self):
        self._test('ipython', bams=self.bams, bed=self.bed4, ipython=True, threads='2')

    def test_08_fastq(self):
        self._test('fastq', fastq=self.fastqs, bwa=self.bwa_path, bed=self.bed4)

    def test_09_debug_and_reuse(self):
        self._test('debug_and_reuse', bams=self.bams, bed=self.bed4, debug=True, reuse_intermediate=True)

    def test_10_reuse_output(self):
        self._test('reuse_output_dir', bams=self.bams, bed=self.bed4, reuse_output_dir=True)

    def test_11_full_hg19(self):
        raise SkipTest

    def test_12_full_hg38(self):
        raise SkipTest

    def test_13_api(self):
        import targqc
        import ngs_utils.reference_data as ref
        from ngs_utils.file_utils import safe_mkdir
        from ngs_utils.parallel import ParallelCfg

        genome = 'hg19-chr21'
        fai_fpath = ref.get_fai(genome)
        output_dirname = 'api'
        output_dir = join(self.results_dir, output_dirname)
        work_dir = join(output_dir, 'work')
        samples = sorted([targqc.Sample(s.name,
             dirpath=safe_mkdir(join(output_dir, s.name)),
             work_dir=safe_mkdir(join(work_dir, s.name)),
             bam=join(self.syn3_dir, s.bam))
                   for s in self.samples], key=lambda _s: _s.key_to_sort())
        parallel_cfg = ParallelCfg(None, None, None, 1, None)
        info('-' * 100)
        targqc.start_targqc(work_dir, output_dir, samples, self.bed4,
                            parallel_cfg, self.bwa_path,
                            fai_fpath=fai_fpath,
                            genome=genome)
        info('-' * 100)
        info('')

        info()
        self._check_results(output_dir, self.samples)



from ngs_utils.logger import warn

metric_names = [
    'Reference size',
    'Regions size/percentage of reference (on target)',
    'Regions size/percentage of reference (on target) %',
    'Coverage Mean',
    'Coverage Mean (on target)',
    'Coverage Standard Deviation',
    'Coverage Standard Deviation (on target)',
    'Reference size',
    'Number of reads',
    'Mapped reads',
    'Mapped reads %',
    'Unmapped reads',
    'Unmapped reads %',
    'Mapped reads (on target)',
    'Mapped reads (on target) %',
    'Mapped paired reads',
    'Mapped paired reads %',
    'Paired reads',
    'Paired reads %',
    'Duplicated reads (flagged)',
    'Duplicated reads (flagged) %',
    'Duplicated reads (flagged) (on target)',
    'Duplicated reads (flagged) (on target) %',
    'Read min length',
    'Read max length',
    'Read mean length',
    'Mean Mapping Quality (on target)',
    'Mismatches (on target)',
    'Insertions (on target)',
    'Deletions (on target)',
    'Homopolymer indels (on target)',
    'Mean Mapping Quality',
    'Mismatches',
    'Insertions',
    'Deletions',
    'Homopolymer indels',
]

ALLOWED_UNITS = ['%']


def parse_qualimap_sample_report(report_fpath):
    value_by_metric = dict()

    def __get_td_tag_contents(line):
        ## examples:
        # <td class=column1>Paired reads</td>
        # <td class=column2>80,244 / 99.89%</td>
        crop_left = line.split('>')
        if len(crop_left) < 2:
            return None
        crop_right = crop_left[1].split('<')
        return crop_right[0].strip()

    def __fill_record(metric_name, line):
        val = __get_td_tag_contents(line)
        val = val.replace(' ', '').replace(',', '')
        val = val.replace(b'\xc2\xa0', '')
        if metric_name == 'Read min/max/mean length':  # special case
            for metric_infix, value in zip(['min', 'max', 'mean'], val.split('/')):
                value_by_metric['Read ' + metric_infix + ' length'] = value
        else:
            if metric_name not in metric_names:
                # warn('Qualimap metric "' + metric_name + '" is not in allowed metric_names')
                return

            num_chars = []
            unit_chars = []
            i = 0
            while i < len(val) and (val[i].isdigit() or val[i] in ['.']):
                num_chars += val[i]
                i += 1
            while i < len(val):
                unit_chars += val[i]
                i += 1
            val_num = ''.join(num_chars)
            val_unit = ''.join(unit_chars)

            if val_unit and val_unit in ALLOWED_UNITS:
                # metric.unit = val_unit
                pass
            try:
                val = int(val_num)
                if val_unit == '%':
                    val = float(val) / 100
            except ValueError:
                try:
                    val = float(val_num)
                    if val_unit == '%':
                        val /= 100
                except ValueError:  # it is a string
                    val = val_num + val_unit

            value_by_metric[metric_name] = val

            if val_unit.startswith('/'):  # for values like "80,220 / 99.86%"
                meta_val = val_unit.replace('/', '').strip()
                if '%' in meta_val:
                    try:
                        val = float(meta_val.replace('%', '')) / 100.0
                    except ValueError:
                        pass
                    else:
                        value_by_metric[metric_name + ' %'] = val

    sections = [['start',                             'Summary'],
                ['globals (on target)',               'Globals (inside of regions)'],
                ['globals',                           'Globals'],
                ['coverage (on target)',              'Coverage (inside of regions)'],
                ['coverage',                          'Coverage'],
                ['mq (on target)',                    'Mapping Quality (inside of regions)'],
                ['mq',                                'Mapping Quality'],
                ['mismatches and indels (on target)', 'Mismatches and indels (inside of regions)'],
                ['mismatches and indels',             'Mismatches and indels'],
                ['finish',                            'Coverage across reference']]  # plots are starting from this line
    on_target_stats_suffix = ' (on target)'
    coverage_stats_prefix = 'Coverage '
    with open(report_fpath) as f:
        cur_section = None
        cur_metric_name = None
        for line in f:
            if 'mapped' in line.lower():
                pass

            if 'class=table-summary' in line:
                cur_section = None
                continue

            if cur_section is None:
                for name, pattern in sections:
                    if pattern in line:
                        cur_section = name
                        break
                if cur_section is None:
                    continue

            if cur_section == 'finish':
                break

            if line.find('class=column1') != -1:
                cur_metric_name = __get_td_tag_contents(line)

                if cur_section.endswith('(on target)'):
                    cur_metric_name += on_target_stats_suffix

                if cur_section.startswith('coverage'):
                    cur_metric_name = coverage_stats_prefix + cur_metric_name

                # if not metric_storage.get_metric(cur_metric_name):  # special case for Duplication rate and Clipped reads (Qualimap v.1 and v.2 difference)
                #     if metric_storage.get_metric(cur_metric_name + on_target_stats_suffix):  # extra 'on target' metrics
                #         cur_metric_name += on_target_stats_suffix

            if cur_metric_name and line.find('class=column2') != -1:
                __fill_record(cur_metric_name, line)
                cur_metric_name = None

    return value_by_metric



# coding=utf-8
import os
from collections import OrderedDict
from os.path import join, abspath, realpath, dirname, relpath
from pybedtools import BedTool

from ensembl import get_merged_cds
from ngs_utils import reference_data, logger
from ngs_utils.bed_utils import get_padded_bed_file, intersect_bed, calc_sum_of_regions, count_bed_cols,\
    calc_bases_within_threshs, calc_rate_within_normal
from ngs_utils.sambamba import index_bam, number_mapped_reads_on_target, number_of_mapped_reads, sambamba_depth
from ngs_utils.file_utils import intermediate_fname, verify_file, safe_mkdir, can_reuse
from ngs_utils.logger import critical, info, err, warn, debug
from ngs_utils.reporting.reporting import ReportSection, Metric, MetricStorage, SampleReport

from targqc.qualimap import report_parser, runner


def get_header_metric_storage(depth_threshs, is_wgs=False, padding=None):
    sections = [
        ReportSection('reads', 'Reads', [
            Metric('Original reads',                       short_name='Orig reads',   multiqc=dict(order=1, kind='reads', min=0)),
            Metric('Reads',                                short_name='Reads',        multiqc=dict(order=1, kind='reads', min=0)),
            Metric('Mapped reads',                         short_name='Mapped',       multiqc=dict(title='Mapped reads', hidden=True, order=2, kind='reads', min=0),  ok_threshold='Percentage of mapped reads', bottom=0, description='samtools view -c -F 4'),
            Metric('Percentage of mapped reads',           short_name='%',            multiqc=dict(title='Mapped', order=3, kind='reads'),                  unit='%', ok_threshold=0.98, bottom=0),
            Metric('Properly paired mapped reads percent', short_name='Paired',       multiqc=dict(order=4, kind='reads'),                                  unit='%', ok_threshold=0.9,  bottom=0, description='Pecent of properly paired mapped reads.'),
            Metric('Duplication rate',                     short_name='Dups',         multiqc=dict(order=5, kind='reads'),                                  unit='%', quality='Less is better',    description='Percent of mapped reads (-F 4), marked as duplicates (-f 1024)'),
            Metric('Read min length',                      short_name='Min len',      multiqc=dict(title='Read min len', hidden=True, kind='reads', min=0), unit='bp',                             description='Read minimum length', is_hidden=True),
            Metric('Read mean length',                     short_name='Avg read len', multiqc=dict(title='Read avg len', order=22, kind='reads', min=0),    unit='bp',                             description='Read average length'),
            Metric('Read max length',                      short_name='Max len',      multiqc=dict(title='Read max len', order=23, kind='reads', min=0),    unit='bp',                             description='Read maximum length'),
            Metric('Median insert size',                   short_name='IS',           multiqc=dict(title='Med. IS', order=21, kind='reads', min=0),         unit='bp',                             description='Median insert size'),
            Metric('Median GC',                            short_name='GC',           multiqc=dict(title='Med. GC', order=21, kind='other'),                unit='%',                              description='Med. GC-content', ),
        ]),
    ]
    if not is_wgs:
        sections.extend([
            ReportSection('target_metrics', 'Target coverage', [
                Metric('Covered bases in target',                         short_name='Trg covered',                         multiqc=dict(hidden=True, title='Trg cov bases', kind='cov'),   unit='bp',                          description='Target bases covered by at least 1 read'),
                Metric('Percentage of target covered by at least 1 read', short_name='%',                                   multiqc=dict(title='Trg cov', order=14, kind='cov'),            unit='%'),
                Metric('Percentage of reads mapped on target',            short_name='Reads on trg',                        multiqc=dict(title='On trg', order=10, kind='trg'),             unit='%',                           description='Percentage of unique mapped reads overlapping target at least by 1 base'),
                Metric('Percentage of reads mapped off target',           short_name='Reads off trg',                       multiqc=dict(hidden=True, title='Off trg', order=11, kind='trg'),            unit='%', quality='Less is better', description='Percentage of unique mapped reads that don\'t overlap target even by 1 base'),
                Metric('Percentage of reads mapped on padded target',     short_name='On trg &#177;' + str(padding) + 'bp', multiqc=dict(order=12, kind='trg'), unit='%',                                                       description='Percentage of reads that overlap target at least by 1 base. Should be 1-2% higher.'),
                Metric('Percentage of usable reads',                      short_name='Usable reads',                        multiqc=dict(order=13, title='Usable', kind='trg'),             unit='%',                           description='Share of unique reads mapped on target in the total number of original reads (reported in the very first column Reads'),
                Metric('Read bases mapped on target',                     short_name='Read bp on trg',                      multiqc=dict(hidden=True, kind='cov'),                          unit='bp'),
            ]),
        ])
    else:
        sections.extend([
            ReportSection('target_metrics_wgs', 'Genome coverage', [
                Metric('Covered bases in genome',                         short_name='Genome covered', multiqc=dict(title='Genome cov bases', hidden=True, kind='cov'),     unit='bp',                          description='Genome bases covered by at least 1 read'),
                Metric('Percentage of genome covered by at least 1 read', short_name='%',              multiqc=dict(title='Genome cov', hidden=True, order=13, kind='cov'), unit='%'),
                Metric('Covered bases in exome',                          short_name='CDS covered',    multiqc=dict(title='CDS cov bases', hidden=True, kind='cov'),        unit='bp',                          description='Covered CDS bases. CDS coordinates are taken from RefSeq'),
                Metric('Percentage of exome covered by at least 1 read',  short_name='%',              multiqc=dict(title='CDS cov', order=14, kind='cov'),                 unit='%',                           description='Percentage of CDS covered by at least 1 read. CDS coordinates are taken from RefSeq'),
                Metric('Percentage of reads mapped on exome',             short_name='Reads on CDS',   multiqc=dict(title='On CDS', order=10, kind='trg'),                  unit='%',                           description='Percentage of reads mapped on CDS. CDS coordinates are taken from RefSeq'),
                Metric('Percentage of reads mapped off exome',            short_name='Off CDS',        multiqc=dict(order=11, kind='trg'),                                  unit='%', quality='Less is better', description='Percentage of reads mapped outside of CDS. CDS coordinates are taken from RefSeq'),
                Metric('Percentage of usable reads',                      short_name='Usable reads',   multiqc=dict(order=12, title='Usable', kind='trg'),                  unit='%',                           description='Share of mapped unique reads in all reads (reported in the very first column Reads)'),
            ]),
        ])

    trg_name = 'target' if not is_wgs else 'genome'
    depth_section = ReportSection('section_name', ('Target' if not is_wgs else 'Genome') + ' coverage depth', [
        Metric('Median ' + trg_name + ' coverage depth',                  short_name='Median',         multiqc=dict(title='Med depth', order=6, kind='cov', min=0)),
        Metric('Mean ' + trg_name + ' coverage depth',                    short_name='Mean',           multiqc=dict(title='Mean depth', order=7, kind='cov', min=0)),
        Metric('Std. dev. of ' + trg_name + ' coverage depth',            short_name='Std dev',        multiqc=dict(order=8, kind='cov', min=0),                                quality='Less is better'),
        Metric('Percentage of ' + trg_name + ' within 20% of med depth',  short_name='&#177;20% med',  multiqc=dict(order=9, kind='cov'),                             unit='%')
    ])
    for depth in depth_threshs:
        name = 'Part of ' + trg_name + ' covered at least by ' + str(depth) + 'x'
        depth_section.add_metric(Metric(name,                             short_name=str(depth) + 'x', multiqc=dict(hidden=True, kind='cov'),                         unit='%', description=name))
    depth_section.add_metric(
        Metric('Estimated ' + trg_name + ' full coverage depth',          short_name='Est full mean',  multiqc=dict(title='Est mean depth', order=7, kind='cov', min=0), description='Estimated mean coverage of full dataset. Calculated as (the total number of raw reads * downsampled mapped reads fraction / total downsampled mapped reads) * downsampled average coverage'),
    )
    sections.append(depth_section)

    sections.append(
        ReportSection('other', 'Other stats' + ('' if is_wgs else ' within the target'), [
            Metric('Mean Mapping Quality',  short_name='Mean MQ',            multiqc=dict(kind='other', min=0),                         description='Mean mapping quality, inside of regions'),
            Metric('Mismatches',            short_name='Mismatches',         multiqc=dict(hidden=True, kind='other', min=0),            description='Mismatches, inside of regions',                       quality='Less is better'),
            Metric('Insertions',            short_name='Insertions',         multiqc=dict(hidden=True, kind='other', min=0),            description='Insertions, inside of regions',                       quality='Less is better'),
            Metric('Deletions',             short_name='Deletions',          multiqc=dict(hidden=True, kind='other', min=0),            description='Deletions, inside of regions',                        quality='Less is better'),
            Metric('Homopolymer indels',    short_name='Homopolymer indels', multiqc=dict(hidden=True, kind='other', min=0),            description='Percentage of homopolymer indels, inside of regions', quality='Less is better'),
            Metric('Qualimap',              short_name='Qualimap report',    multiqc=dict(hidden=True, kind='other'),                   description='Qualimap report'),
            Metric('Sex',                   short_name='sex',                multiqc=dict(order=20,    kind='other'), is_hidden=True),
        ])
    )

    ms = MetricStorage(
        general_section=ReportSection('general_section', '', [
            Metric('Target',                                     short_name='Target',                            common=True),
            Metric('Ready target (clean, sorted and annotated)', short_name='Ready target',                      common=True),
            Metric('Reference size',                             short_name='Reference bp', unit='bp',           common=True),
            Metric('Regions in target',                          short_name='Regions in target',                 common=True),
            Metric('Bases in target',                            short_name='Target bp', unit='bp',              common=True),
            Metric('Percentage of reference',                    short_name='Percentage of reference', unit='%', common=True),
            Metric('Genes in target',                            short_name='Genes in target',                   common=True),
            Metric('Scope',                                      short_name='Scope',                             common=True, is_hidden=True),
        ]),
        sections=sections
    )
    return ms


# def _dedup_and_flag_stat(cnf, bam_fpath):
#     bam_stats = flag_stat(cnf, bam_fpath)
#     info('Total reads: ' + Metric.format_value(bam_stats['total']))
#     info('Total mapped reads: ' + Metric.format_value(bam_stats['mapped']))
#     info('Total dup reads: ' + Metric.format_value(bam_stats['duplicates']))
#     info('Total properly paired reads: ' + Metric.format_value(bam_stats['properly paired']))
#
#     # dedup_bam_dirpath = join(cnf.work_dir, source.dedup_bam)
#     # safe_mkdir(dedup_bam_dirpath)
#     # dedup_bam_fpath = join(dedup_bam_dirpath, add_suffix(basename(bam_fpath), source.dedup_bam))
#     # remove_dups(cnf, bam_fpath, output_fpath=dedup_bam_fpath, use_grid=False)
#     # dedup_bam_stats = samtools_flag_stat(cnf, dedup_bam_fpath)
#     # info('Total reads after dedup (samtools view -F 1024): ' + Metric.format_value(dedup_bam_stats['total']))
#     # info('Total mapped reads after dedup (samtools view -F 1024): ' + Metric.format_value(dedup_bam_stats['mapped']))
#     return bam_stats  # dedup_bam_fpath, bam_stats, dedup_bam_stats


def parse_qualimap_insert_size(qualimap_insert_size_fpath):
    d = dict()
    zero_insertsize = 0
    with open(qualimap_insert_size_fpath) as f:
        for l in f:
            if not l.strip() or l.startswith('#'):
                continue
            insertsize, count = l.split(None, 1)
            insertsize = int(round(float(insertsize)))
            count = float(count) / 1000000
            if insertsize == 0:
                zero_insertsize = count
            else:
                d[insertsize] = count

    # Find median without importing anything to do it for us
    num_counts = sum(d.values())
    cum_counts = 0
    median_insert_size = None
    for thisins, thiscount in d.items():
        cum_counts += thiscount
        if cum_counts >= num_counts/2:
            median_insert_size = thisins
            break
    return median_insert_size


def parse_qualimap_gc_content(qualimap_gc_content_fpath):
    avg_gc = 0
    avg_human_gc = 0
    with open(qualimap_gc_content_fpath) as f:
        for l in f:
            if l.startswith('#'):
                continue
            sections = l.split(None, 3)
            gc = int(round(float(sections[0])))         # %
            content = float(sections[1]) / 100.0        # share of reads with this GC %
            avg_gc += gc * content
            if len(sections) > 2:
                human_content = float(sections[2]) / 100.0  # share of human reference with this GC %
                avg_human_gc += gc * human_content
    return avg_gc, avg_human_gc


def parse_qualimap_coverage_hist(qualimap_coverage_hist_fpath):
    bases_by_depth = OrderedDict()
    with open(qualimap_coverage_hist_fpath) as f:
        for l in f:
            if l.startswith('#'):
                pass
            else:
                cov, bases = map(int, map(float, l.strip().split()))
                bases_by_depth[cov] = bases

    # calculating median coverage
    num_counts = sum(bases_by_depth.values())
    cum_counts = 0
    median_coverage = None
    for thiscov, thiscount in bases_by_depth.items():
        cum_counts += thiscount
        if cum_counts >= num_counts/2:
            median_coverage = thiscov
            break

    return bases_by_depth, median_coverage


def parse_qualimap_results(sample):
    if not verify_file(sample.qualimap_html_fpath):
        critical('QualiMap report was not found')

    qualimap_value_by_metric = report_parser.parse_qualimap_sample_report(sample.qualimap_html_fpath)
    bases_by_depth, median_depth = parse_qualimap_coverage_hist(sample.qualimap_cov_hist_fpath)
    median_gc, median_human_gc = parse_qualimap_gc_content(sample.qualimap_gc_hist_fpath)
    median_ins_size = parse_qualimap_insert_size(sample.qualimap_ins_size_hist_fpath)

    def find_rec(name, percent=False, on_target=True):
        if on_target:
            name_on_target = name + ' (on target)'
            if percent:
                name_on_target += ' %'
            res = qualimap_value_by_metric.get(name_on_target)
            if res:
                return res
        if percent:
            name += ' %'
        return qualimap_value_by_metric.get(name)

    depth_stats = dict(
        ave_depth       = find_rec('Coverage Mean'),
        stddev_depth    = find_rec('Coverage Standard Deviation'),
        median_depth    = median_depth,
        bases_by_depth  = bases_by_depth
    )
    target_stats = dict(
        reference_size  = find_rec('Reference size'),
        target_size     = find_rec('Regions size/percentage of reference'),
        target_fraction = find_rec('Regions size/percentage of reference', percent=True),
    )
    reads_stats = dict(
        total                    = find_rec('Number of reads'),
        mapped                   = find_rec('Mapped reads', on_target=False),
        mapped_rate              = find_rec('Mapped reads', percent=True, on_target=False),
        unmapped                 = find_rec('Unmapped reads'),
        unmapped_rate            = find_rec('Unmapped reads', percent=True),
        mapped_on_target         = find_rec('Mapped reads (on target)'),
        mapped_rate_on_target    = find_rec('Mapped reads (on target)', percent=True),
        mapped_paired            = find_rec('Mapped paired reads', on_target=False),
        mapped_paired_rate       = find_rec('Mapped paired reads', percent=True, on_target=False),
        paired                   = find_rec('Paired reads'),
        paired_rate              = find_rec('Paired reads', percent=True),
        dup                      = find_rec('Duplicated reads (flagged)'),
        dup_rate                 = find_rec('Duplicated reads (flagged)', percent=True),
        min_len                  = find_rec('Read min length'),
        max_len                  = find_rec('Read max length'),
        ave_len                  = find_rec('Read mean length'),
        median_gc                = median_gc,
        median_human_gc          = median_human_gc,
        median_ins_size          = median_ins_size,
    )
    indels_stats = dict(
        mean_mq     = find_rec('Mean Mapping Quality'),
        mismatches  = find_rec('Mismatches'),
        insertions  = find_rec('Insertions'),
        deletions   = find_rec('Deletions'),
        homo_indels = find_rec('Homopolymer indels'),
    )
    return depth_stats, reads_stats, indels_stats, target_stats


chry_key_regions_by_genome = {
    'hg19': join(dirname(abspath(__file__)), 'gender', 'chrY.hg19.bed'),
    'hg38': join(dirname(abspath(__file__)), 'gender', 'chrY.hg38.bed'),
}
MALE_TARGET_REGIONS_FACTOR = 0.7
AVE_DEPTH_THRESHOLD_TO_DETERMINE_SEX = 5
FEMALE_Y_COVERAGE_FACTOR = 10.0

def determine_sex(work_dir, bam_fpath, ave_depth, genome, target_bed=None):
    info()
    info('Determining sex')

    male_bed = None
    for k in chry_key_regions_by_genome:
        if k in genome:
            male_bed = BedTool(chry_key_regions_by_genome.get(k))
            break
    if not male_bed:
        warn('Warning: no male key regions for ' + genome + ', cannot identify sex')
        return None

    male_area_size = male_bed.count()
    info('Male region total size: ' + str(male_area_size))

    if target_bed:
        male_bed = BedTool(target_bed).intersect(male_bed).merge()
        target_male_area_size = male_bed.count()
        if target_male_area_size < male_area_size * MALE_TARGET_REGIONS_FACTOR:
            info('Target male region total size is ' + str(target_male_area_size) + ', which is less than the ' +
                 'checked male regions size * ' + str(MALE_TARGET_REGIONS_FACTOR) +
                 ' (' + str(male_area_size * MALE_TARGET_REGIONS_FACTOR) + ') - cannot determine sex')
            return None
        else:
            info('Target male region total size is ' + str(target_male_area_size) + ', which is higher than the ' +
                 'checked male regions size * ' + str(MALE_TARGET_REGIONS_FACTOR) +
                 ' (' + str(male_area_size * MALE_TARGET_REGIONS_FACTOR) + '). ' +
                 'Determining sex based on coverage in those regions.')
    else:
        info('WGS, determining sex based on chrY key regions coverage.')

    info('Detecting sex by comparing the Y chromosome key regions coverage and average coverage depth.')
    if not bam_fpath:
        critical('BAM file is required.')
    index_bam(bam_fpath)

    chry_cov_output_fpath = sambamba_depth(work_dir, male_bed, bam_fpath, [])
    chry_mean_coverage = get_mean_cov(chry_cov_output_fpath)
    info('Y key regions average depth: ' + str(chry_mean_coverage))
    ave_depth = float(ave_depth)
    info('Sample average depth: ' + str(ave_depth))
    if ave_depth < AVE_DEPTH_THRESHOLD_TO_DETERMINE_SEX:
        info('Sample average depth is too low (less then ' + str(AVE_DEPTH_THRESHOLD_TO_DETERMINE_SEX) +
             ') - cannot determine sex')
        return None

    if chry_mean_coverage == 0:
        info('Y depth is 0 - it\s female')
        sex = 'F'
    else:
        factor = ave_depth / chry_mean_coverage
        info('Sample depth / Y depth = ' + str(factor))
        if factor > FEMALE_Y_COVERAGE_FACTOR:  # if mean target coverage much higher than chrY coverage
            info('Sample depth is more than ' + str(FEMALE_Y_COVERAGE_FACTOR) + ' times higher than Y depth - it\s female')
            sex = 'F'
        else:
            info('Sample depth is not more than ' + str(FEMALE_Y_COVERAGE_FACTOR) + ' times higher than Y depth - it\s male')
            sex = 'M'
    info('Sex is ' + sex)
    info()
    return sex


def get_mean_cov(bedcov_output_fpath):
    mean_cov = []
    mean_cov_col = None
    total_len = 0
    with open(bedcov_output_fpath) as bedcov_file:
        for line in bedcov_file:
            if line.startswith('#'):
                mean_cov_col = line.split('\t').index('meanCoverage')
                continue
            line_tokens = line.replace('\n', '').split()
            start, end = map(int, line_tokens[1:3])
            size = end - start
            mean_cov.append(float(line_tokens[mean_cov_col]) * size)
            total_len += size
    mean_cov = sum(mean_cov) / total_len if total_len > 0 else 0
    return mean_cov


def _qualimap_outputs(sample):
    return [v for k, v in sample.__dict__.iteritems() if k.startswith('qualimap_') and k.endswith('_fpath')]


def make_general_reports(view, samples, target, genome, depth_threshs, bed_padding,
                         num_pairs_by_sample=None, reuse=False, is_debug=False, reannotate=False, fai_fpath=None):
    if all(all(can_reuse(fp, [s.bam, target.qualimap_bed_fpath] if target.bed else s.bam)
               for fp in _qualimap_outputs(s))
           for s in samples):
        debug('All QualiMap files for all samples exist and newer than BAMs and BEDs, reusing')
    else:
        info('Running QualiMap...')
        view.run(runner.run_qualimap,
            [[s.work_dir, s.qualimap_dirpath, _qualimap_outputs(s), s.bam, genome, target.qualimap_bed_fpath, view.cores_per_job]
             for s in samples])

        for s in samples:
            for fp in _qualimap_outputs(s):
                verify_file(fp, is_critical=True)

    summary_reports = []

    for sample in samples:
        info('-'*70)
        info(sample.name)
        debug('-'*70)
        debug('Parsing QualiMap results...')
        depth_stats, reads_stats, indels_stats, target_stats = parse_qualimap_results(sample)

        _prep_report_data(sample, depth_stats, reads_stats, indels_stats, target_stats,
                          target, num_pairs_by_sample, genome, depth_threshs, fai_fpath=fai_fpath)

        r = _build_report(depth_stats, reads_stats, indels_stats, sample, target,
                          depth_threshs, bed_padding, sample_num=len(samples), is_debug=is_debug,
                          reannotate=reannotate)
        summary_reports.append(r)

    return summary_reports


def _prep_report_data(sample, depth_stats, reads_stats, indels_stats, target_stats,
                      target, num_pairs_by_sample, genome, depth_threshs, fai_fpath=None):
    sample.avg_depth = depth_stats['ave_depth']

    if num_pairs_by_sample and sample.name in num_pairs_by_sample:
        reads_stats['original_num_reads'] = num_pairs_by_sample[sample.name] * 2

    chrom_lengths = reference_data.get_chrom_lengths(genome=genome, fai_fpath=fai_fpath)
    if 'Y' in chrom_lengths or 'chrY' in chrom_lengths:
        reads_stats['gender'] = determine_sex(sample.work_dir, sample.bam, depth_stats['ave_depth'],
                                              genome, target.get_capture_bed())
        info()

    if 'bases_by_depth' in depth_stats:
        depth_stats['bases_within_threshs'], depth_stats['rates_within_threshs'] = calc_bases_within_threshs(
            depth_stats['bases_by_depth'],
            target_stats['target_size'] if not target.is_wgs else target_stats['reference_size'],
            depth_threshs)

        if depth_stats['median_depth'] > 0:
            depth_stats['wn_20_percent'] = calc_rate_within_normal(
                depth_stats['bases_by_depth'],
                depth_stats['median_depth'],
                target_stats['target_size'] if not target.is_wgs else target_stats['reference_size'])

    if target_stats['target_size']:
        target.bases_num = target_stats['target_size']
        target.fraction  = target_stats['target_fraction']
    else:
        target.bases_num = target_stats['reference_size']

    reads_stats['mapped_dedup'] = number_of_mapped_reads(sample.work_dir, sample.bam, dedup=True)

    if not target.is_wgs:
        reads_stats['mapped_dedup_on_target'] = number_mapped_reads_on_target(
            sample.work_dir, target.get_capture_bed().cut(range(3)).saveas().fn, sample.bam, dedup=True, target_name='target') or 0

        reads_stats['mapped_dedup_on_padded_target'] = number_mapped_reads_on_target(
            sample.work_dir, target.padded_bed_fpath, sample.bam, dedup=True, target_name='padded_target') or 0

    else:
        cds_bed = get_merged_cds(genome)
        info('Using the CDS reference BED to calc "reads on CDS"')
        reads_stats['mapped_dedup_on_exome'] = number_mapped_reads_on_target(
            sample.work_dir, cds_bed, sample.bam, dedup=True, target_name='exome') or 0

    return depth_stats, reads_stats, indels_stats


def _build_report(depth_stats, reads_stats, mm_indels_stats, sample, target,
                  depth_threshs, bed_padding, sample_num, is_debug=False, reannotate=False):
    report = SampleReport(sample, metric_storage=get_header_metric_storage(depth_threshs, is_wgs=target.bed_fpath is None, padding=bed_padding))

    def _add(_metric_name, _val, url=None):
        return report.add_record(_metric_name, _val, silent=(sample_num > 1 and not is_debug), url=url)

    _add('Qualimap', 'Qualimap', url=relpath(sample.qualimap_html_fpath, sample.dirpath))
    if reads_stats.get('gender') is not None:
        _add('Sex', reads_stats['gender'])

    debug('* General coverage statistics *')
    _add('Reads', reads_stats['total'])
    _add('Mapped reads', reads_stats['mapped'])
    # _add('Unmapped reads', reads_stats['totaAvgl'] - reads_stats['mapped'])
    percent_mapped = 1.0 * (reads_stats['mapped'] or 0) / reads_stats['total'] if reads_stats['total'] else None
    assert percent_mapped <= 1.0 or percent_mapped is None, str(percent_mapped)
    _add('Percentage of mapped reads', percent_mapped)
    # percent_unmapped = 1.0 * (reads_stats['total'] - reads_stats['mapped']) / reads_stats['total'] if reads_stats['total'] else None
    # assert percent_unmapped <= 1.0 or percent_unmapped is None, str(percent_unmapped)
    # _add('Percentage of unmapped reads', percent_unmapped)
    if reads_stats.get('mapped_paired') is not None:
        total_paired_reads_pecent = 1.0 * (reads_stats['mapped_paired'] or 0) / reads_stats['total'] if reads_stats['total'] else None
        assert total_paired_reads_pecent <= 1.0 or total_paired_reads_pecent is None, str(total_paired_reads_pecent)
        _add('Properly paired mapped reads percent', total_paired_reads_pecent)
    # if reads_stats.get('paired') is not None:
    #     total_paired_reads_pecent = 1.0 * (reads_stats['paired'] or 0) / reads_stats['total'] if reads_stats['total'] else None
    #     assert total_paired_reads_pecent <= 1.0 or total_paired_reads_pecent is None, str(total_paired_reads_pecent)
    #     _add('Properly paired reads percent', total_paired_reads_pecent)
    # if dedup_bam_stats:
    # dup_rate = 1 - (1.0 * dedup_bam_stats['mapped'] / bam_stats['mapped']) if bam_stats['mapped'] else None
    _add('Duplication rate', reads_stats['dup_rate'])
    # _add('Dedupped mapped reads', reads_stats['mapped'] - reads_stats[''])
    _add('Median GC', reads_stats['median_gc'])
    _add('Median insert size', reads_stats['median_ins_size'])

    debug()

    if not target.is_wgs:
        debug('* Target coverage statistics *')
        if target.original_bed_fpath:
            _add('Target', target.original_bed_fpath)
            if count_bed_cols(target.original_bed_fpath) == 3 or reannotate:
                _add('Ready target (clean, sorted and annotated)', target.capture_bed_fpath)
        else:
            _add('Target', target.capture_bed_fpath)
        _add('Bases in target', target.bases_num)
        _add('Percentage of reference', target.fraction)
        _add('Regions in target', target.regions_num)
        _add('Scope', 'targeted')
        _add('Genes in target', len(target.gene_keys_list))
    else:
        debug('* Genome coverage statistics *')
        _add('Target', 'whole genome')
        _add('Reference size', target.bases_num)
        _add('Scope', 'WGS')

    trg_type = 'target' if not target.is_wgs else 'genome'

    if 'bases_within_threshs' in depth_stats:
        bases_within_threshs = depth_stats['bases_within_threshs']
        v_covered_bases_in_targ = bases_within_threshs.items()[0][1]
        v_percent_covered_bases_in_targ = 1.0 * (v_covered_bases_in_targ or 0) / target.bases_num if target.bases_num else None
        assert v_percent_covered_bases_in_targ <= 1.0 or v_percent_covered_bases_in_targ is None, str(v_percent_covered_bases_in_targ)

        _add('Covered bases in ' + trg_type, v_covered_bases_in_targ)
        _add('Percentage of ' + trg_type + ' covered by at least 1 read', v_percent_covered_bases_in_targ)

    if not target.is_wgs:
        debug('Getting number of mapped reads on target...')
        # mapped_reads_on_target = number_mapped_reads_on_target(cnf, target_info.bed, bam_fpath)
        if 'mapped_dedup_on_target' in reads_stats:
            # _add('Reads mapped on target', reads_stats['mapped_on_target'])
            debug('Unique mapped reads on target: ' + str(reads_stats['mapped_dedup_on_target']))
            percent_mapped_dedup_on_target = 1.0 * reads_stats['mapped_dedup_on_target'] / reads_stats['mapped_dedup'] if reads_stats['mapped_dedup'] != 0 else None
            _add('Percentage of reads mapped on target', percent_mapped_dedup_on_target)
            assert percent_mapped_dedup_on_target <= 1.0 or percent_mapped_dedup_on_target is None, str(percent_mapped_dedup_on_target)

            percent_mapped_dedup_off_target = 1.0 * (reads_stats['mapped_dedup'] - reads_stats['mapped_dedup_on_target']) / reads_stats['mapped_dedup'] if reads_stats['mapped_dedup'] != 0 else None
            _add('Percentage of reads mapped off target', percent_mapped_dedup_off_target)
            assert percent_mapped_dedup_off_target <= 1.0 or percent_mapped_dedup_off_target is None, str(percent_mapped_dedup_off_target)

            percent_usable = 1.0 * reads_stats['mapped_dedup_on_target'] / reads_stats['total'] if reads_stats['total'] != 0 else None
            _add('Percentage of usable reads', percent_usable)
            assert percent_usable <= 1.0 or percent_usable is None, str(percent_usable)

        read_bases_on_targ = int(target.bases_num * depth_stats['ave_depth'])  # sum of all coverages
        _add('Read bases mapped on target', read_bases_on_targ)

        if 'mapped_dedup_on_padded_target' in reads_stats:
            # _add('Reads mapped on padded target', reads_stats['mapped_reads_on_padded_target'])
            percent_mapped_on_padded_target = 1.0 * reads_stats['mapped_dedup_on_padded_target'] / reads_stats['mapped_dedup'] if reads_stats['mapped_dedup'] else None
            _add('Percentage of reads mapped on padded target', percent_mapped_on_padded_target)
            assert percent_mapped_on_padded_target <= 1.0 or percent_mapped_on_padded_target is None, str(percent_mapped_on_padded_target)

    elif 'mapped_dedup_on_exome' in reads_stats:
        # _add('Reads mapped on target', reads_stats['mapped_on_target'])
        percent_mapped_on_exome = 1.0 * reads_stats['mapped_dedup_on_exome'] / reads_stats['mapped_dedup'] if reads_stats['mapped_dedup'] != 0 else None
        if percent_mapped_on_exome:
            _add('Percentage of reads mapped on exome', percent_mapped_on_exome)
            assert percent_mapped_on_exome <= 1.0 or percent_mapped_on_exome is None, str(percent_mapped_on_exome)
            percent_mapped_off_exome = 1.0 - percent_mapped_on_exome
            _add('Percentage of reads mapped off exome ', percent_mapped_off_exome)

        percent_usable = 1.0 * reads_stats['mapped_dedup'] / reads_stats['total'] if reads_stats['total'] != 0 else None
        _add('Percentage of usable reads', percent_usable)
        assert percent_usable <= 1.0 or percent_usable is None, str(percent_usable)

    debug()
    _add('Mean ' + trg_type + ' coverage depth', depth_stats['ave_depth'])
    if 'original_num_reads' in reads_stats:
        _add('Original reads', reads_stats['original_num_reads'])
        times_downsampled = 1.0 * reads_stats['original_num_reads'] / reads_stats['total']
        est_full_cov = times_downsampled * depth_stats['ave_depth']
        _add('Estimated ' + trg_type + ' full coverage depth', est_full_cov)
    _add('Median ' + trg_type + ' coverage depth', depth_stats['median_depth'])
    if depth_stats['median_depth'] > 0:
        _add('Std. dev. of ' + trg_type + ' coverage depth', depth_stats['stddev_depth'])
    # _add('Minimal ' + trg_type + ' coverage depth', depth_stats['min_depth'])
    # _add('Maximum ' + trg_type + ' coverage depth', depth_stats['max_depth'])
    if 'wn_20_percent' in depth_stats:
        _add('Percentage of ' + trg_type + ' within 20% of med depth', depth_stats['wn_20_percent'])
        assert depth_stats['wn_20_percent'] <= 1.0 or depth_stats['wn_20_percent'] is None, str(depth_stats['wn_20_percent'])

    if 'bases_within_threshs' in depth_stats:
        for depth, bases in depth_stats['bases_within_threshs'].items():
            fraction_val = 1.0 * (bases or 0) / target.bases_num if target.bases_num else 0
            if fraction_val > 0:
                _add('Part of ' + trg_type + ' covered at least by ' + str(depth) + 'x', fraction_val)
            assert fraction_val <= 1.0 or fraction_val is None, str(fraction_val)
    debug()

    _add('Read mean length', reads_stats['ave_len'])
    _add('Read min length', reads_stats['min_len'])
    _add('Read max length', reads_stats['max_len'])
    _add('Mean Mapping Quality', mm_indels_stats['mean_mq'])
    _add('Mismatches', mm_indels_stats['mismatches'])
    _add('Insertions', mm_indels_stats['insertions'])
    _add('Deletions', mm_indels_stats['deletions'])
    _add('Homopolymer indels', mm_indels_stats['homo_indels'])

    debug()
    info('Saving reports...')
    report.save_json(sample.targqc_json_fpath)
    report.save_txt(sample.targqc_txt_fpath)
    report.save_html(sample.targqc_html_fpath, caption='Target coverage statistics for ' + sample.name)
    debug()
    debug('Saved to ' + dirname(report.txt_fpath))
    return report


def get_total_reads_number_from_fastqc(sample, fastqc_dirpath):
    fastqc_txt_fpaths = find_fastqc_txt(sample, fastqc_dirpath)
    if not fastqc_txt_fpaths:
        return
    num_reads = 0
    for fpath in fastqc_txt_fpaths:
        with open(fpath) as f_in:
            for line in f_in:
                if 'total sequences' in line.lower():
                    num_reads += int(line.strip().split('\t')[-1])
                    break
    return num_reads


def find_fastqc_txt(sample_name, fastqc_dirpath):
    l_fastqc_dirpath = join(fastqc_dirpath, sample_name + '_R1_fastqc')
    r_fastqc_dirpath = join(fastqc_dirpath, sample_name + '_R2_fastqc')
    fastqc_txt_fpaths = [join(l_fastqc_dirpath, 'fastqc_data.txt'), join(r_fastqc_dirpath, 'fastqc_data.txt')]
    if all(os.path.isfile(fpath) for fpath in fastqc_txt_fpaths):
        return fastqc_txt_fpaths
    else:
        return None




# coding=utf-8

from collections import defaultdict
from os.path import isfile, join

from ngs_utils import reference_data
from ngs_utils.bed_utils import count_bed_cols
from ngs_utils.sambamba import sambamba_depth
from ngs_utils.call_process import run
from ngs_utils.file_utils import intermediate_fname, verify_file, file_transaction, can_reuse
from ngs_utils.logger import info, debug
from ngs_utils.utils import OrderedDefaultDict

import ensembl as ebl


def make_region_reports(view, work_dir, samples, target, genome, depth_thresholds):
    bed_fpath = target.bed_fpath or target.wgs_bed_fpath

    if all(can_reuse(s.targqc_region_tsv, [s.bam, bed_fpath]) for s in samples):
        debug('All region reports exist, reusing')
        return [s.targqc_region_tsv for s in samples]

    info('Calculating coverage statistics for CDS and exon regions from RefSeq...')

    depth_thresholds_by_sample = dict()
    for s in samples:
        depth_thresholds_by_sample[s.name] = depth_thresholds

    debug()
    debug('Running sambamba...')
    sambamba_depth_output_fpaths = view.run(sambamba_depth,
        [[s.work_dir, bed_fpath, s.bam, depth_thresholds_by_sample[s.name], None, s.name]
         for s in samples])
    assert len(sambamba_depth_output_fpaths) == len(samples), \
        'Number of sambamba results = ' + str(len(sambamba_depth_output_fpaths)) + \
        ' which is less then the number of samples ' + str(len(samples))

    debug()
    debug('Parsing sambamba results and writing results...')
    view.run(_proc_sambamba_depth,
        [[sambamba_output_fpath, s.targqc_region_tsv, s.name, depth_thresholds_by_sample[s.name]]
         for sambamba_output_fpath, s in zip(sambamba_depth_output_fpaths, samples)])

    info('Done.')
    return [s.targqc_region_tsv for s in samples]


def _proc_sambamba_depth(sambamba_depth_output_fpath, output_fpath, sample_name, depth_thresholds):
    read_count_col = None
    mean_cov_col = None
    median_cov_col = None
    min_depth_col = None
    std_dev_col = None
    wn_20_pcnt_col = None

    regions_by_genekey = defaultdict(list)
    #####################################
    #####################################
    if can_reuse(output_fpath, sambamba_depth_output_fpath):
        return output_fpath

    debug('Reading coverage statistics and writing regions to ' + output_fpath)

    def write_line(f, fields):
        f.write('\t'.join(fields) + '\n')

    with file_transaction(None, output_fpath) as tx:
        with open(sambamba_depth_output_fpath) as sambabma_depth_file, open(tx, 'w') as out:
            total_regions_count = 0
            for line in sambabma_depth_file:
                fs = line.strip('\n').split('\t')
                if line.startswith('#'):
                    fs = line.split('\t')
                    read_count_col = fs.index('readCount')
                    mean_cov_col = fs.index('meanCoverage')
                    median_cov_col = fs.index('medianCoverage') if 'medianCoverage' in fs else None
                    min_depth_col = fs.index('minDepth') if 'minDepth' in fs else None
                    std_dev_col = fs.index('stdDev') if 'stdDev' in fs else None
                    wn_20_pcnt_col = fs.index('percentWithin20PercentOfMedian') if 'percentWithin20PercentOfMedian' in fs else None

                    write_line(out, [
                        '#Chr',
                        'Start',
                        'End',
                        'Size',
                        'Gene',
                        'Exon',
                        'Strand',
                        'Feature',
                        'Biotype',
                        'Transcript',
                        'Tx overlap',
                        'Exome overlap',
                        'Min depth',
                        'Avg depth',
                        'Median depth',
                        'Std dev',
                        'W/n 20% of median',
                    ] + ['{}x'.format(ths) for ths in depth_thresholds])
                    continue

                chrom = fs[0]
                start, end = int(fs[1]), int(fs[2])
                region_size = end - start
                gene_name = fs[ebl.BedCols.GENE] if read_count_col != ebl.BedCols.GENE else '.'
                exon = fs[ebl.BedCols.EXON]
                strand = fs[ebl.BedCols.STRAND]
                feature = fs[ebl.BedCols.FEATURE]
                biotype = fs[ebl.BedCols.BIOTYPE]
                transcript = fs[ebl.BedCols.ENSEMBL_ID]
                transcript_overlap = fs[ebl.BedCols.TX_OVERLAP_PERCENTAGE]
                exome_overlap = fs[ebl.BedCols.EXON_OVERLAPS_PERCENTAGE]
                avg_depth = float(fs[mean_cov_col])
                min_depth = int(fs[min_depth_col]) if min_depth_col is not None else '.'
                std_dev = float(fs[std_dev_col]) if std_dev_col is not None else '.'
                median_depth = int(fs[median_cov_col]) if median_cov_col is not None else '.'
                rate_within_normal = float(fs[wn_20_pcnt_col]) if wn_20_pcnt_col is not None else '.'
                last_cov_col = max(mean_cov_col, median_cov_col, std_dev_col, wn_20_pcnt_col)
                rates_within_threshs = fs[last_cov_col + 1:-1]

                write_line(out, map(str, [
                        chrom,
                        start,
                        end,
                        region_size,
                        gene_name,
                        exon,
                        strand,
                        feature,
                        biotype,
                        transcript,
                        ((transcript_overlap + '%') if transcript_overlap not in ['', None, '.'] else '.'),
                        ((exome_overlap + '%') if exome_overlap not in ['', None, '.'] else '.'),
                        min_depth,
                        avg_depth,
                        median_depth,
                        std_dev,
                        rate_within_normal,
                    ] + rates_within_threshs))

                total_regions_count += 1
                if total_regions_count > 0 and total_regions_count % 10000 == 0:
                    debug('  Processed {0:,} regions'.format(total_regions_count))
        debug('Total regions: ' + str(len(regions_by_genekey)))
    return output_fpath

# def _get_values_from_row(fields, cols):
#     return [fields[col] if col else None for col in cols]


# def _parse_report(sample, gene_by_name_and_chrom):
#     rep = PerRegionSampleReport()
#     rep.txt_fpath = sample.targqc_region_txt
#     rep.tsv_fpath = sample.targqc_region_tsv
#
#     with open(sample.targqc_region_tsv) as f:
#         min_depth_col = None
#         median_depth_col = None
#         avg_depth_col = None
#         std_dev_col = None
#         wn_20_pcnt_col = None
#         for l in f:
#             rep.rows.append(1)
#             fs = l.strip().split('\t')
#
#             if l.startswith('#'):
#                 if len(fs) > 9:
#                     min_depth_col = fs.index('Min depth') if 'Min depth' in fs else min_depth_col
#                     median_depth_col = fs.index('Median depth') if 'Median depth' in fs else median_depth_col
#                     avg_depth_col = fs.index('Ave depth') if 'Ave depth' in fs else avg_depth_col
#                     std_dev_col = fs.index('Std dev') if 'Std dev' in fs else std_dev_col
#                     wn_20_pcnt_col = fs.index('W/n 20% of median depth') if 'W/n 20% of median depth' in fs else wn_20_pcnt_col
#                 continue
#             chrom, start, end, size, gene_name, strand, feature, biotype, transcript_id = fs[:9]
#             min_depth, median_depth, avg_depth, std_dev, wn20ofmedian = \
#                 _get_values_from_row(fs, [min_depth_col, median_depth_col, avg_depth_col, std_dev_col, wn_20_pcnt_col])
#
#             if gene_name not in ('.', '', None):
#                 region = Region(gene_name=gene_name, transcript_id=transcript_id, exon_num=None,
#                                 strand=strand, biotype=biotype, feature=feature, extra_fields=list(), chrom=chrom,
#                                 start=int(start) if start not in ('.', '', None) else None,
#                                 end=int(end) if end not in ('.', '', None) else None,
#                                 size=int(size) if size not in ('.', '', None) else None,
#                                 min_depth=float(min_depth) if min_depth not in ('.', '', None) else None,
#                                 median_depth=float(median_depth) if median_depth not in ('.', '', None) else None,
#                                 avg_depth=float(avg_depth) if avg_depth not in ('.', '', None) else None,
#                                 std_dev=float(std_dev) if std_dev not in ('.', '', None) else None,
#                                 rate_within_normal=float(wn20ofmedian) if wn20ofmedian and wn20ofmedian not in ('.', '', None) else None, )
#                 region.sample_name = gene_by_name_and_chrom[(gene_name, chrom)].sample_name
#                 depth_thresholds = cfg.depth_thresholds
#                 rates_within_threshs = OrderedDict((depth, None) for depth in depth_thresholds)
#                 rates = fs[-(len(depth_thresholds)):]
#                 for i, t in enumerate(rates_within_threshs):
#                     rates_within_threshs[t] = float(rates[i]) if rates[i] not in ('.', '', None) else None
#                 region.rates_within_threshs = rates_within_threshs
#                 if 'Capture' in feature:
#                     gene_by_name_and_chrom[(gene_name, chrom)].add_amplicon(region)
#                 elif 'CDS' in feature or feature == 'Exon':
#                     gene_by_name_and_chrom[(gene_name, chrom)].add_exon(region)
#                 else:
#                     gene_by_name_and_chrom[(gene_name, chrom)].chrom = region.chrom
#                     gene_by_name_and_chrom[(gene_name, chrom)].strand = region.strand
#                     gene_by_name_and_chrom[(gene_name, chrom)].avg_depth = region.avg_depth
#                     gene_by_name_and_chrom[(gene_name, chrom)].min_depth = region.min_depth
#                     gene_by_name_and_chrom[(gene_name, chrom)].median_depth = region.median_depth
#                     gene_by_name_and_chrom[(gene_name, chrom)].rates_within_threshs = region.rates_within_threshs
#     return rep


# def get_detailed_metric_storage(depth_threshs):
#     return MetricStorage(
#         general_section=ReportSection(metrics=[
#             Metric('Sample'),
#         ]),
#         sections=[ReportSection(metrics=[
#             Metric('Chr'),
#             Metric('Start'),
#             Metric('End'),
#             Metric('Size'),
#             Metric('Gene'),
#             Metric('Strand'),
#             Metric('Feature'),
#             Metric('Biotype'),
#             Metric('Transcript'),
#             Metric('Min depth'),
#             Metric('Median depth'),
#             Metric('Ave depth'),
#             Metric('Std dev', description='Coverage depth standard deviation'),
#             Metric('W/n 20% of median depth', description='Percentage of the region that lies within 20% of an median depth.', unit='%'),
#             # Metric('Norm depth', description='Ave region depth devided by median depth of sample'),
#         ] + [
#             Metric('{}x'.format(thresh), description='Bases covered by at least {} reads'.format(thresh), unit='%')
#             for thresh in depth_threshs
#         ])]
#     )


# def _sambamba_depth_to_regions(sambamba_depth_output_fpath, sample_name, target_type, depth_thresholds):
#     read_count_col = None
#     mean_cov_col = None
#     median_cov_col = None
#     min_depth_col = None
#     std_dev_col = None
#     wn_20_pcnt_col = None
#
#     regions_by_genekey = defaultdict(list)
#     #####################################
#     #####################################
#     debug('Reading coverage statistics...')
#     with open(sambamba_depth_output_fpath) as sambabma_depth_file:
#         total_regions_count = 0
#         for line in sambabma_depth_file:
#             fs = line.replace('\n', '').split('\t')
#             if line.startswith('#'):
#                 fs = line.split('\t')
#                 read_count_col = fs.index('readCount')
#                 mean_cov_col = fs.index('meanCoverage')
#                 median_cov_col = fs.index('medianCoverage') if 'medianCoverage' in fs else None
#                 min_depth_col = fs.index('minDepth') if 'minDepth' in fs else None
#                 std_dev_col = fs.index('stdDev') if 'stdDev' in fs else None
#                 wn_20_pcnt_col = fs.index('W/n 20% of median depth') if 'W/n 20% of median depth' in fs else None
#                 continue
#             chrom = fs[0]
#             start, end = map(int, fs[1:3])
#             region_size = end - start
#             gene_name = fs[3] if read_count_col != 3 else None
#             ave_depth = float(fs[mean_cov_col])
#             min_depth = int(fs[min_depth_col]) if min_depth_col is not None else '.'
#             std_dev = float(fs[std_dev_col]) if std_dev_col is not None else '.'
#             median_depth = int(fs[median_cov_col]) if median_cov_col is not None else '.'
#             rate_within_normal = float(fs[wn_20_pcnt_col]) if wn_20_pcnt_col is not None else '.'
#             last_cov_col = max(mean_cov_col, median_cov_col, std_dev_col, wn_20_pcnt_col)
#             rates_within_threshs = fs[last_cov_col + 1:-1]
#
#             extra_fields = tuple(fs[4:read_count_col]) if read_count_col > 4 else ()
#
#             region = Region(
#                 sample_name=sample_name, chrom=chrom,
#                 start=start, end=end, size=region_size,
#                 avg_depth=ave_depth, median_depth=median_depth,
#                 gene_name=gene_name, extra_fields=extra_fields,
#                 rate_within_normal=rate_within_normal)
#             regions_by_genekey[(gene_name, chrom)].append(region)
#
#             region.rates_within_threshs = OrderedDict((depth, float(rate) / 100.0) for (depth, rate) in zip(depth_thresholds, rates_within_threshs))
#             region.min_depth = min_depth
#             region.std_dev = std_dev
#
#             if target_type == 'amplicons':
#                 region.feature = 'Capture'
#             else:
#                 if extra_fields:
#                     region.exon_num = extra_fields[0]
#                     if len(extra_fields) >= 2:
#                         region.strand = extra_fields[1]
#                     if len(extra_fields) >= 3:
#                         region.feature = extra_fields[2]
#                     else:
#                         region.feature = 'CDS'
#                     if len(extra_fields) >= 4:
#                         region.biotype = extra_fields[3]
#                     if len(extra_fields) >= 5:
#                         region.transcript_id = extra_fields[4]
#
#             total_regions_count += 1
#             if total_regions_count > 0 and total_regions_count % 10000 == 0:
#                 debug('  Processed {0:,} regions'.format(total_regions_count))
#     debug('Total regions: ' + str(len(regions_by_genekey)))
#
#     # #####################################
#     # #####################################
#     # info('Second round of sambamba depth - calculating depth within 20% bounds of average depth')
#     # sambamba_depth_output_fpath = sambamba_depth(cnf, bed, bam, depth_thresholds=[int()])
#     # if not sambamba_depth_output_fpath:
#     #     continue
#     #
#     # info('Adding rates within normal...')
#     # with open(sambamba_depth_output_fpath) as sambabma_depth_file:
#     #     total_regions_count = 0
#     #     for region, line in zip(regions, (l for l in sambabma_depth_file if not l.startswith('#'))):
#     #         line_tokens = line.replace('\n', '').split()
#     #         rate_within_low_bound = line_tokens[std_dev_col + 1]
#     #         rate_within_higher_bound = line_tokens[std_dev_col + 2]
#     #         regions.rate_within_normal = rate_within_low_bound - rate_within_higher_bound
#     #
#     #         total_regions_count += 1
#     #         if total_regions_count > 0 and total_regions_count % 10000 == 0:
#     #              info('  Processed {0:,} regions'.format(total_regions_count))
#     #     info('Processed {0:,} regions'.format(total_regions_count))
#
#     return regions_by_genekey


# def _write_regions(refseq_regions_by_genekey, target_regions_by_genekey, genekeys_list,
#                    output_fpath, depth_thresholds, reuse=False):
#     if reuse and isfile(output_fpath) and verify_file(output_fpath):
#         debug('Regional report ' + output_fpath + ' exists, reusing')
#         return
#
#     debug('Writing regions to ' + output_fpath)
#     with file_transaction(None, output_fpath) as tx:
#         with open(tx, 'w') as out:
#             out.write('#' + '\t'.join([
#                 'Chr',
#                 'Start',
#                 'End',
#                 'Size',
#                 'Gene',
#                 'Strand',
#                 'Feature',
#                 'Biotype',
#                 'Transcript',
#                 'Min depth',
#                 'Median depth',
#                 'Std dev',
#                 'W/n 20% of median',
#             ] + ['{}x'.format(ths) for ths in depth_thresholds]) + '\n')
#
#             for (gene_name, chrom) in genekeys_list:
#                 for r in refseq_regions_by_genekey[(gene_name, chrom)] + \
#                          target_regions_by_genekey[(gene_name, chrom)]:
#                     out.write('\t'.join(map(str, [
#                         r.chrom,
#                         r.start,
#                         r.end,
#                         r.get_size(),
#                         r.gene_name,
#                         r.strand or '.',
#                         r.feature,
#                         r.biotype or '.',
#                         r.transcript_id or '.',
#                         r.min_depth,
#                         r.median_depth,
#                         r.avg_depth,
#                         r.std_dev,
#                         r.rate_within_normal,
#                     ] + [r.rates_within_threshs.get(ths) for ths in depth_thresholds])) + '\n')


# def _prev_generate_regional_reports_from_bam(view, work_dir, samples, target, features_bed):
#     # if reuse and verify_file(sample.targqc_region_tsv, silent=True):
#     #     debug(sample.targqc_region_tsv + ' exists, reusing')
#     #     return _parse_report(sample, target)
#
#     #####################################
#     #####################################
#     ready_to_report_genes = []
#     ready_to_report_set = set()
#
#     # debug('Interating in parallel')
#     # target_and_exons_fpath = join(work_dir, 'target_and_exons.bed')
#     # amplicons_f = open(target_bed) if target_bed else None
#     # exons_f = open(features_bed) if features_bed else None
#     #
#     # with open(target_and_exons_fpath, 'w') as out:
#     #     for cur_gene in gene_by_name_and_chrom.values():
#     #         amplicon_l = None
#     #         exon_l = None
#     #
#     #         if amplicon_l:  # first line after which the previous "while True" was interrupted - this line belongs to the next gene
#     #             assert amplicon_l.split('\t')[3] == cur_gene.gene_name
#     #             out.write(amplicon_l)
#     #
#     #         # iterate over regions until meat another gene
#     #         if amplicons_f:
#     #             while True:
#     #                 amplicon_l = next(amplicons_f, None)
#     #                 if not amplicon_l: break
#     #                 if amplicon_l.startswith('#'): continue
#     #                 chrom, _, _, gname = amplicon_l.strip().split('\t')[:4]
#     #                 if (gname, chrom) != (cur_gene.gene_name, cur_gene.chrom) and gname != '.': break  # another gene
#     #                 out.write(amplicon_l)
#     #
#     #         if exon_l:  # first line after which the previous "while True" was interrupted - this line belongs to the next gene
#     #             assert exon_l.split('\t')[3] == cur_gene.gene_name
#     #             out.write(exon_l)
#     #
#     #         if exons_f and cur_gene.gene_name not in genes_not_in_refseq:  # gene will be found!
#     #             while True:
#     #                 exon_l = next(exons_f, None)
#     #                 if not exon_l: break
#     #                 if exon_l.startswith('#'): continue
#     #                 chrom, _, _, gname, _, _, feature = exon_l.strip().split('\t')[:7]
#     #                 if (gname, chrom) != (cur_gene.gene_name, cur_gene.chrom): break
#     #                 if feature in ['CDS', 'Exon']:
#     #                     out.write(exon_l)
#     #
#     # if target_bed: amplicons_f.close()
#     # if features_bed: exons_f.close()
#
#     # debug('Saved mixed BED into ' + target_and_exons_fpath)
#     # sys.exit(1)
#
#     gene_by_name_and_chrom = build_gene_objects_list(features_bed, target.gene_keys_list)
#
#     for (bed_fpath, target_type) in zip([target.bed_fpath, exons_and_cds_features], ['amplicons', 'exons']):  # features are canonical
#         if not bed_fpath:
#             continue
#         info()
#         info('Calculating coverage statistics for ' + ('CDS and ncRNA exons...' if target_type == 'exons' else 'the regions in the target BED file...'))
#         sambamba_depth_output_fpath = sambamba_depth(work_dir, bed_fpath, sample.bam, depth_thresholds, reuse=cfg.reuse_intermediate)
#         regions = _sambamba_depth_to_regions(sambamba_depth_output_fpath, sample.name, target_type, depth_thresholds)
#
#         #####################################
#         #####################################
#         info('Filling regions stats into the gene objects...')
#         cur_unannotated_gene = None
#         total_regions_count = 0
#         for region in regions:
#             if region.feature == 'Capture':
#                 if region.gene_name != '.':
#                     cur_unannotated_gene = None
#                     gene = gene_by_name_and_chrom[(region.gene_name, region.chrom)]
#                     if (gene.gene_name, gene.chrom) not in ready_to_report_set:
#                         ready_to_report_genes.append(gene)
#                         ready_to_report_set.add((gene.gene_name, gene.chrom))
#                     gene.add_amplicon(region)
#                 else:
#                     if cur_unannotated_gene is None:
#                         cur_unannotated_gene = GeneInfo(gene_name=region.gene_name, chrom=region.chrom, feature='NotAnnotatedSummary')
#                         ready_to_report_genes.append(cur_unannotated_gene)
#                     cur_unannotated_gene.add_amplicon(region)
#
#             else:
#                 gene = gene_by_name_and_chrom[(region.gene_name, region.chrom)]
#                 gene.add_exon(region)
#                 if not target.bed_fpath:  # in case if only reporting based on features_bed
#                     if (gene.gene_name, gene.chrom) not in ready_to_report_set:
#                         ready_to_report_genes.append(gene)
#                         ready_to_report_set.add((gene.gene_name, gene.chrom))
#
#             # row = [region.chrom, region.start, region.end, region.get_size(), region.gene_name, region.strand,
#             #        region.feature, region.biotype, region.transcript_id, region.min_depth, region.avg_depth, region.std_dev,
#             #        region.rate_within_normal]
#             # row = [Metric.format_value(val, human_readable=True) for val in row]
#             # rates = [Metric.format_value(val, unit='%', human_readable=True) for val in region.rates_within_threshs.values()]
#             # row.extend(rates)
#             # col_widths = [max(len(v), w) for v, w in izip(row, col_widths)]
#
#             total_regions_count += 1
#             if total_regions_count > 0 and total_regions_count % 10000 == 0:
#                 info('  Processed {0:,} regions'.format(total_regions_count))
#         info('Processed {0:,} regions'.format(total_regions_count))
#
#     #####################################
#     #####################################
#     report = PerRegionSampleReport(sample=sample, metric_storage=get_detailed_metric_storage(depth_thresholds))
#     # report.add_record('Sample', sample.name)
#     report.txt_fpath = sample.targqc_region_txt
#     report.tsv_fpath = sample.targqc_region_tsv
#
#     debug('Arranging regions to report...')
#     regions = []
#     for g in ready_to_report_genes:
#         for a in g.get_amplicons():
#             regions.append(a)
#         for e in g.get_exons():
#             regions.append(e)
#         if g.get_exons():
#             process_gene(g, depth_thresholds)
#             regions.append(g)
#
#     debug('Preparting report rows...')
#     for reg in regions:
#         r = report.add_row()
#         r.add_record('Chr', reg.chrom)
#         r.add_record('Start', reg.start)
#         r.add_record('End', reg.end)
#         r.add_record('Size', reg.get_size())
#         r.add_record('Gene', reg.gene_name)
#         r.add_record('Strand', reg.strand)
#         r.add_record('Feature', reg.feature)
#         r.add_record('Biotype', reg.biotype)
#         r.add_record('Transcript', reg.transcript_id)
#         r.add_record('Min depth', reg.min_depth)
#         r.add_record('Median depth', reg.median_depth)
#         r.add_record('Ave depth', reg.avg_depth)
#         r.add_record('Std dev', reg.std_dev)
#         r.add_record('W/n 20% of median depth', reg.rate_within_normal)
#         for ths in depth_thresholds:
#             r.add_record('{}x'.format(ths), reg.rates_within_threshs.get(ths) if reg.rates_within_threshs else None)
#
#     safe_mkdir(dirname(report.tsv_fpath))
#     debug('Flattening TSV records...')
#     header_rows, flat_rows = report.flatten(None, human_readable=False)
#     debug('Writing TSV...')
#     write_tsv_rows((header_rows, flat_rows), report.tsv_fpath)
#     debug('Flattening TXT records...')
#     header_rows, flat_rows = report.flatten(None, human_readable=True)
#     debug('Writing TXT...')
#     write_txt_rows((header_rows, flat_rows), report.txt_fpath)
#
#     # un_annotated_summary_region = next((g for g in gene_by_name_and_chrom.values() if g.gene_name == '.'), None)
#     # if un_annotated_summary_region and un_annotated_amplicons:
#     #     un_annotated_summary_region.feature = 'NotAnnotatedSummary'
#     #     for ampl in un_annotated_amplicons:
#     #         ampl.gene_name = un_annotated_summary_region.gene_name
#     #         un_annotated_summary_region.add_amplicon(ampl)
#     #         add_region_to_report(report, ampl, depth_thresholds)
#     #     add_region_to_report(report, un_annotated_summary_region, depth_thresholds)
#
#     # report.save_txt(sample.targqc_region_txt)
#     # report.save_tsv(sample.targqc_region_tsv)
#     info('Regions (total ' + str(len(report.rows)) + ') saved into:')
#     info('  ' + report.txt_fpath)
#     return report


# def process_gene(gene, depth_thresholds):
#     gene.rates_within_threshs = OrderedDict((depth, None) for depth in depth_thresholds)
#     if gene.size == 0:
#         gene.start = None
#         gene.end = None
#         return
#     exons = gene.get_exons()
#     if not exons:
#         return
#
#     total_depth = sum(e.avg_depth * e.size for e in exons)
#     gene.size = sum(e.size for e in exons)
#     gene.avg_depth = total_depth / gene.size
#     if exons[0].std_dev not in ['.', '', None]:
#         sum_of_sq_var = sum((((e.avg_depth - e.std_dev) - gene.avg_depth) ** 2 + ((e.avg_depth + e.std_dev) - gene.avg_depth) ** 2) * e.size for e in exons)
#         gene.std_dev = math.sqrt(sum_of_sq_var / 2 / float(gene.size))
#     for t in depth_thresholds:
#         total_rate = sum(e.rates_within_threshs[t] * e.size for e in exons)
#         rate = total_rate / gene.size
#         gene.rates_within_threshs[t] = rate


# def add_region_to_report(report, region, depth_threshs, file_to_write=None, col_widths=None):
#     if file_to_write:
#         report.rows.append(1)
#         rep_region = Row(parent_report=report)
#     else:
#         rep_region = report.add_row()
#     rep_region.add_record('Chr', region.chrom)
#     rep_region.add_record('Start', region.start)
#     rep_region.add_record('End', region.end)
#     rep_region.add_record('Size', region.get_size())
#     rep_region.add_record('Gene', region.gene_name)
#     rep_region.add_record('Strand', region.strand)
#     rep_region.add_record('Feature', region.feature)
#     rep_region.add_record('Biotype', region.biotype)
#     rep_region.add_record('Transcript', region.transcript_id)
#     rep_region.add_record('Min depth', region.min_depth)
#     rep_region.add_record('Ave depth', region.avg_depth)
#     rep_region.add_record('Std dev', region.std_dev)
#     rep_region.add_record('W/n 20% of ave depth', region.rate_within_normal)
#
#     if region.rates_within_threshs is None:
#         warn('Error: no rates_within_threshs for ' + str(region))
#     for thresh in depth_threshs:
#         rep_region.add_record('{}x'.format(thresh), region.rates_within_threshs.get(thresh) if region.rates_within_threshs else None)
#
#     if file_to_write:
#         for fpath in fpaths_to_write:
#             human_readable = fpath.endswith('txt')
#             flat_row = []
#             for m in report.metric_storage.get_metrics(None, skip_general_section=True):
#                 rec = BaseReport.find_record(rep_region.records, m.name)
#                 if rec:
#                     flat_row.append(rec.format(human_readable=human_readable))
#
#             with open(fpath, 'a') as out:
#                 if fpath.endswith('tsv'):
#                     out.write('\t'.join([val for val in flat_row]) + '\n')
#                 else:
#                     col_widths = col_widths or repeat(0)
#                     for val, w in izip(flat_row, col_widths):
#                         out.write(val + (' ' * (w - len(val) + 2)))
#                     out.write('\n')
#
#     return col_widths

# def _bases_by_depth(depth_vals, depth_thresholds):
#     bases_by_min_depth = {depth: 0 for depth in depth_thresholds}
#
#     for depth_value in depth_vals:
#         for threshold in depth_thresholds:
#             if depth_value >= threshold:
#                 bases_by_min_depth[threshold] += 1
#
#         return [1.0 * bases_by_min_depth[thres] / len(depth_vals) if depth_vals else 0
#                 for thres in depth_thresholds]





import os
import random
import gzip
from itertools import izip, product
from os.path import splitext, dirname, join, basename, isfile, abspath

from ngs_utils import sambamba
from ngs_utils.bam_utils import verify_bam
from ngs_utils.call_process import run
from ngs_utils.file_utils import open_gzipsafe, file_transaction, verify_file, add_suffix, safe_mkdir, which, can_reuse
from ngs_utils.logger import critical, debug, info, warn, err


def make_downsampled_fpath(work_dir, fastq_fpath):
    return join(work_dir, add_suffix(basename(fastq_fpath), 'subset'))

def make_pair_counts_fpath(work_dir):
    return join(work_dir, 'full_pairs_count.txt')

def make_bam_fpath(work_dir):
    return join(work_dir, 'downsampled.bam')


def proc_fastq(samples, parall_view, work_dir, bwa_prefix, downsample_to, num_pairs_by_sample=None, dedup=True):
    num_pairs_by_sample = num_pairs_by_sample or dict()
    if downsample_to:
        # Read pairs counts
        debug()
        if all(s.name in num_pairs_by_sample for s in samples):
            debug('Using read pairs counts extracted from FastQC reports')
        elif all(can_reuse(make_pair_counts_fpath(s.work_dir), s.l_fpath) for s in samples):
            debug('Reusing pairs counts, reading from files')
            num_pairs_by_sample = {s.name: int(open(make_pair_counts_fpath(s.work_dir)).read().strip()) for s in samples}
        else:
            info('Counting read pairs')
            num_pairs = parall_view.run(count_read_pairs, [[s.name, s.work_dir, s.l_fpath] for s in samples])
            num_pairs_by_sample = {s.name: pairs_count for s, pairs_count in zip(samples, num_pairs)}

        # Downsampling
        debug()
        if all(can_reuse(make_downsampled_fpath(s.work_dir, s.l_fpath), s.l_fpath) and
               can_reuse(make_downsampled_fpath(s.work_dir, s.r_fpath), s.r_fpath) for s in samples):
            debug('Reusing downsampled FastQ')
            for s in samples:
                s.l_fpath = make_downsampled_fpath(s.work_dir, s.l_fpath)
                s.r_fpath = make_downsampled_fpath(s.work_dir, s.r_fpath)
        else:
            if isinstance(downsample_to, float):
                info('Downsampling FastQ to ' + str(float(downsample_to)) + ' fraction of reads')
            else:
                info('Downsampling FastQ to ' + str(int(downsample_to)) + ' read pairs')
            fastq_pairs = parall_view.run(downsample,
                [[s.work_dir, s.name, s.l_fpath, s.r_fpath, downsample_to, num_pairs_by_sample.get(s.name)]
                 for s in samples])
            for s, (l_r, r_r) in zip(samples, fastq_pairs):
                s.l_fpath = l_r
                s.r_fpath = r_r
    else:
        info('Skipping downsampling')

    debug()
    if all(can_reuse(make_bam_fpath(s.work_dir), [s.l_fpath, s.r_fpath]) for s in samples):
        debug('All downsampled BAM exists, reusing')
        for s in samples:
            s.bam = make_bam_fpath(s.work_dir)
    else:
        bwa = which('bwa')
        if not isfile(bwa):
            critical('BWA not found under ' + bwa)
        smb = sambamba.get_executable()
        if not (bwa and smb):
            if not bwa:         err('Error: bwa is required for the alignment pipeline')
            if not smb:         err('Error: sambamba is required for the alignment pipeline')
            critical('Tools required for alignment not found')
        info('Aligning reads to the reference')
        bam_fpaths = parall_view.run(align,
            [[s.work_dir, s.name, s.l_fpath, s.r_fpath, bwa, smb, bwa_prefix, dedup, parall_view.cores_per_job]
             for s in samples])

        bam_fpaths = map(verify_bam, bam_fpaths)
        if len(bam_fpaths) < len(samples):
            critical('Some samples were not aligned successfully.')
        for bam, s in zip(bam_fpaths, samples):
            s.bam = bam

    return num_pairs_by_sample


def count_read_pairs(s_name, work_dir, fastq_fpath):
    from ngs_utils.logger import info

    pairs_counts_fpath = make_pair_counts_fpath(work_dir)
    if can_reuse(pairs_counts_fpath, fastq_fpath):
        with open(pairs_counts_fpath) as f:
            return int(f.read().strip())
    else:
        info('Counting read pairs in ' + s_name + ', writing to ' + pairs_counts_fpath)
        pairs_number = _count_records_in_fastq(fastq_fpath)
        with open(pairs_counts_fpath, 'w') as out:
            out.write(str(pairs_number))
        return pairs_number


def _count_records_in_fastq(fastq_fpath):
    return sum(1 for _ in open_gzipsafe(fastq_fpath)) / 4


def downsample(work_dir, sample_name, fastq_left_fpath, fastq_right_fpath, downsample_to, num_pairs=None):
    """ get N random headers from a fastq file without reading the
    whole thing into memory
    modified from: http://www.biostars.org/p/6544/
    """
    sample_name = sample_name or splitext(''.join(lc if lc == rc else '' for lc, rc in izip(fastq_left_fpath, fastq_right_fpath)))[0]

    l_out_fpath = make_downsampled_fpath(work_dir, fastq_left_fpath)
    r_out_fpath = make_downsampled_fpath(work_dir, fastq_right_fpath)
    if can_reuse(l_out_fpath, [fastq_left_fpath, fastq_right_fpath]):
        return l_out_fpath, r_out_fpath

    info('Processing ' + sample_name)
    if num_pairs is None:
        info(sample_name + ': counting number of reads in fastq...')
        num_pairs = _count_records_in_fastq(fastq_left_fpath)
    if num_pairs > LIMIT:
        info(sample_name + ' the number of reads is higher than ' + str(LIMIT) +
             ', sampling from only first ' + str(LIMIT))
        num_pairs = LIMIT
    info(sample_name + ': ' + str(num_pairs) + ' reads')
    num_downsample_pairs = int(downsample_to * num_pairs) if isinstance(downsample_to, float) else downsample_to
    if num_pairs <= num_downsample_pairs:
        info(sample_name + ': and it is less than ' + str(num_downsample_pairs) + ', so no downsampling.')
        return fastq_left_fpath, fastq_right_fpath
    else:
        info(sample_name + ': downsampling to ' + str(num_downsample_pairs))
        rand_records = sorted(random.sample(xrange(num_pairs), num_downsample_pairs))

    info('Opening ' + fastq_left_fpath)
    fh1 = open_gzipsafe(fastq_left_fpath)
    info('Opening ' + fastq_right_fpath)
    fh2 = open_gzipsafe(fastq_right_fpath) if fastq_right_fpath else None

    out_files = (l_out_fpath, r_out_fpath) if r_out_fpath else (l_out_fpath,)

    written_records = 0
    with file_transaction(work_dir, out_files) as tx_out_files:
        if isinstance(tx_out_files, basestring):
            tx_out_f1 = tx_out_files
        else:
            tx_out_f1, tx_out_f2 = tx_out_files
        info('Opening ' + str(tx_out_f1) + ' to write')
        sub1 = open_gzipsafe(tx_out_f1, "w")
        info('Opening ' + str(tx_out_f2) + ' to write')
        sub2 = open_gzipsafe(tx_out_f2, "w") if r_out_fpath else None
        rec_no = -1
        for rr in rand_records:
            while rec_no < rr:
                rec_no += 1
                for i in range(4): fh1.readline()
                if fh2:
                    for i in range(4): fh2.readline()
            for i in range(4):
                sub1.write(fh1.readline())
                if sub2:
                    sub2.write(fh2.readline())
            written_records += 1
            if written_records % 10000 == 0:
                info(sample_name + ': written ' + str(written_records) + ', rec_no ' + str(rec_no + 1))
            if rec_no > num_pairs:
                info(sample_name + ' reached the limit of ' + str(num_pairs), ' read lines, stopping.')
                break
        info(sample_name + ': done, written ' + str(written_records) + ', rec_no ' + str(rec_no))
        fh1.close()
        sub1.close()
        if fastq_right_fpath:
            fh2.close()
            sub2.close()

    info(sample_name + ': done downsampling, saved to ' + l_out_fpath + ' and ' + r_out_fpath + ', total ' + str(written_records) + ' paired reads written')
    return l_out_fpath, r_out_fpath


def align(work_dir, sample_name, l_fpath, r_fpath, bwa, smb, bwa_prefix, dedup=True, threads=1):
    info('Running bwa to align reads...')
    bam_fpath = make_bam_fpath(work_dir)
    if can_reuse(bam_fpath, [l_fpath, r_fpath]):
        return bam_fpath

    tmp_dirpath = join(work_dir, 'sambamba_tmp_dir')
    safe_mkdir(tmp_dirpath)

    bwa_cmdline = ('{bwa} mem -t {threads} -v 2 {bwa_prefix} {l_fpath} {r_fpath} | ' +
                   '{smb} view /dev/stdin -t {threads} -f bam -S -o - | ' +
                   '{smb} sort /dev/stdin -t {threads} --tmpdir {tmp_dirpath} -o {bam_fpath}').format(**locals())
    run(bwa_cmdline, output_fpath=bam_fpath, stdout_to_outputfile=False)

    if dedup:
        dedup_bam_fpath = add_suffix(bam_fpath, 'dedup')
        dedup_cmdl = '{smb} markdup -t {threads} {bam_fpath} {dedup_bam_fpath}'.format(**locals())
        run(dedup_cmdl, output_fpath=dedup_bam_fpath, stdout_to_outputfile=False)
        verify_bam(dedup_bam_fpath)
        os.rename(dedup_bam_fpath, bam_fpath)

    sambamba.index_bam(bam_fpath)

# samtools view -b -S -u - |
# sambamba sort -N -t 8 -m 682M --tmpdir /Molly/saveliev/cancer-dream-syn3/work/align/syn3-normal/split/tx/tmpwdXndE/syn3-normal-sort-1_20000000-sorttmp-full
    # -o /Molly/saveliev/cancer-dream-syn3/work/align/syn3-normal/split/tx/tmpwdXndE/syn3-normal-sort-1_20000000.bam
    # /dev/stdin

    # if dedup:
    #     info()
    #     info('Calling SamBlaster to mark duplicates')
    #     markdup_sam_fpath = markdup_sam(sam_fpath, samblaster)
    #     if markdup_sam_fpath:
    #         sam_fpath = markdup_sam_fpath
    # info()

    # info('Converting to BAM')
    # cmdline = sambamba.get_executable() + ' view -t {threads} -S -f bam {sam_fpath}'.format(**locals())
    # run(cmdline, output_fpath=bam_fpath, reuse=cfg.reuse_intermediate)
    #
    # info()
    # info('Sorting BAM')
    # prefix = splitext(sorted_bam_fpath)[0]
    # cmdline = sambamba.get_executable() + ' sort -t {threads} {bam_fpath} -o {sorted_bam_fpath}'.format(**locals())
    # run(cmdline, output_fpath=sorted_bam_fpath, stdout_to_outputfile=False, reuse=cfg.reuse_intermediate)

    return bam_fpath


LIMIT = 500*1000*1000


# def markdup_sam(in_sam_fpath, samblaster=None, reuse=False):
#     """Perform non-stream based deduplication of SAM input files using samblaster.
#     """
#     if not samblaster:
#         samblaster = 'samblaster'
#         if which(samblaster) is None:
#             warn('No samblaster, can\'t mark duplicates.')
#             return None
#
#     out_sam_fpath = add_suffix(in_sam_fpath, 'markdup')
#     cmdline = '{samblaster} -i {in_sam_fpath} -o {out_sam_fpath}'.format(**locals())
#     run(cmdline, output_fpath=out_sam_fpath, stdout_to_outputfile=False, reuse=reuse)
#     return out_sam_fpath





from collections import defaultdict
from os.path import isfile, join, basename
from ngs_utils import reference_data
from pybedtools import BedTool
from ngs_utils.bed_utils import sort_bed, verify_bed, get_genes_from_bed
from ngs_utils.file_utils import iterate_file, add_suffix, intermediate_fname, file_transaction, verify_file, can_reuse
from ngs_utils.logger import debug, info, warn
from ngs_utils.utils import OrderedDefaultDict

from bed_annotation.bed_annotation import overlap_with_features, get_sort_key, tx_priority_sort_key
import ensembl as ebl

from targqc import config as cfg


class Target:
    def __init__(self, work_dir, output_dir, fai_fpath, bed_fpath=None,
                 padding=None, reannotate=False, genome=None, is_debug=False):
        self.bed = None
        self.original_bed_fpath = None
        self.bed_fpath = None  # with genomic features
        self.capture_bed_fpath = None  # w/o genomic features
        self.qualimap_bed_fpath = None
        self.padded_bed_fpath = None

        self.gene_keys_set = set()  # set of pairs (gene_name, chrom)
        self.gene_keys_list = list()  # list of pairs (gene_name, chrom)
        self.regions_num = None

        self.bases_num = None
        self.fraction = None

        if bed_fpath:
            debug('Using target BED file ' + bed_fpath)
            self.is_wgs = False
            verify_bed(bed_fpath, is_critical=True)
            self.original_bed_fpath = bed_fpath
            self._make_target_bed(bed_fpath, work_dir, output_dir, padding=padding,
                is_debug=is_debug, fai_fpath=fai_fpath, genome=genome, reannotate=reannotate)
        else:
            debug('No input BED. Assuming whole genome. For region-based reports, analysing RefSeq CDS.')
            self.is_wgs = True
            self._make_wgs_regions_file(work_dir, genome=genome)

    def get_capture_bed(self):
        if not self.is_wgs:
            if self.bed.field_count() > ebl.BedCols.FEATURE:
                return self.bed.filter(lambda x: x[ebl.BedCols.FEATURE] == 'capture')
            else:
                return self.bed
        else:
            return None

    def _make_target_bed(self, bed_fpath, work_dir, output_dir, is_debug,
                         padding=None, fai_fpath=None, genome=None, reannotate=False):
        clean_target_bed_fpath = intermediate_fname(work_dir, bed_fpath, 'clean')
        if not can_reuse(clean_target_bed_fpath, bed_fpath):
            debug()
            debug('Cleaning target BED file...')
            bed = BedTool(bed_fpath)
            if bed.field_count() > 4:
                bed = bed.cut(range(4))
            bed = bed\
                .filter(lambda x: x.chrom and not any(x.chrom.startswith(e) for e in ['#', ' ', 'track', 'browser']))\
                .remove_invalid()
            with file_transaction(work_dir, clean_target_bed_fpath) as tx:
                bed.saveas(tx)
            debug('Saved to ' + clean_target_bed_fpath)
            verify_file(clean_target_bed_fpath, is_critical=True)

        sort_target_bed_fpath = intermediate_fname(work_dir, clean_target_bed_fpath, 'sorted')
        if not can_reuse(sort_target_bed_fpath, clean_target_bed_fpath):
            debug()
            debug('Sorting target BED file...')
            sort_target_bed_fpath = sort_bed(clean_target_bed_fpath, output_bed_fpath=sort_target_bed_fpath, fai_fpath=fai_fpath)
            debug('Saved to ' + sort_target_bed_fpath)
            verify_file(sort_target_bed_fpath, is_critical=True)

        if genome in ebl.SUPPORTED_GENOMES:
            ann_target_bed_fpath = intermediate_fname(work_dir, sort_target_bed_fpath, 'ann_plus_features')
            if not can_reuse(ann_target_bed_fpath, sort_target_bed_fpath):
                debug()
                if BedTool(sort_target_bed_fpath).field_count() == 3 or reannotate:
                        debug('Annotating target BED file and collecting overlapping genome features')
                        overlap_with_features(sort_target_bed_fpath, ann_target_bed_fpath, work_dir=work_dir,
                             genome=genome, extended=True, reannotate=reannotate, only_canonical=True)
                else:
                    debug('Overlapping with genomic features:')
                    overlap_with_features(sort_target_bed_fpath, ann_target_bed_fpath, work_dir=work_dir,
                         genome=genome, extended=True, only_canonical=True)
                debug('Saved to ' + ann_target_bed_fpath)
                verify_file(ann_target_bed_fpath, is_critical=True)
        else:
            ann_target_bed_fpath = sort_target_bed_fpath

        final_clean_target_bed_fpath = intermediate_fname(work_dir, ann_target_bed_fpath, 'clean')
        if not can_reuse(final_clean_target_bed_fpath, ann_target_bed_fpath):
            bed = BedTool(ann_target_bed_fpath).remove_invalid()
            with file_transaction(work_dir, final_clean_target_bed_fpath) as tx:
                bed.saveas(tx)
            verify_file(final_clean_target_bed_fpath, is_critical=True)

        self.bed_fpath = final_clean_target_bed_fpath
        self.bed = BedTool(self.bed_fpath)

        self.capture_bed_fpath = add_suffix(join(output_dir, basename(bed_fpath)), 'clean_sorted_ann')
        with file_transaction(work_dir, self.capture_bed_fpath) as tx:
            self.get_capture_bed().saveas(tx)

        gene_key_set, gene_key_list = get_genes_from_bed(bed_fpath)
        self.gene_keys_set = gene_key_set
        self.gene_keys_list = gene_key_list
        self.regions_num = self.get_capture_bed().count()

        self._make_qualimap_bed(work_dir)
        if padding:
            self._make_padded_bed(work_dir, fai_fpath, padding)

    def _make_padded_bed(self, work_dir, fai_fpath, padding):
        if self.is_wgs:
            return None

        self.padded_bed_fpath = intermediate_fname(work_dir, self.capture_bed_fpath, 'padded')
        if can_reuse(self.padded_bed_fpath, self.capture_bed_fpath):
            return BedTool(self.padded_bed_fpath)

        padded_bed = self.bed.slop(b=padding, g=fai_fpath).sort().merge()
        with file_transaction(work_dir, self.padded_bed_fpath) as tx:
            padded_bed.saveas(tx)
        verify_file(self.padded_bed_fpath, is_critical=True)
        return BedTool(self.padded_bed_fpath)

    def _make_qualimap_bed(self, work_dir):
        if self.is_wgs:
            return None

        self.qualimap_bed_fpath = intermediate_fname(work_dir, self.capture_bed_fpath, 'qualimap_ready')
        if can_reuse(self.qualimap_bed_fpath, self.capture_bed_fpath):
            return self.qualimap_bed_fpath

        debug('Merging and saving BED into required bed6 format for Qualimap')
        bed = self.bed.sort().merge()
        with file_transaction(work_dir, self.qualimap_bed_fpath) as tx:
            with open(tx, 'w') as out:
                for i, region in enumerate(x for x in bed):
                    region = [x for x in list(region) if x]
                    fillers = [str(i), "1.0", "+"]
                    full = region + fillers[:6 - len(region)]
                    out.write("\t".join(full) + "\n")
        verify_file(self.qualimap_bed_fpath, is_critical=True)
        return self.qualimap_bed_fpath

    def _make_wgs_regions_file(self, work_dir, genome=None):
        self.wgs_bed_fpath = join(work_dir, 'targqc_features_to_report.bed')
        if can_reuse(self.wgs_bed_fpath, ebl.ensembl_gtf_fpath(genome)):
            return self.wgs_bed_fpath

        chr_order = reference_data.get_chrom_order(genome or cfg.genome)

        r_by_tx_by_gene = OrderedDefaultDict(lambda: defaultdict(list))
        all_features = ebl.get_all_features(genome or cfg.genome, high_confidence=True)

        debug('Select best transcript to report')
        for r in all_features:
            if r[ebl.BedCols.FEATURE] != 'gene':
                gene = r[ebl.BedCols.HUGO]
                tx = r[ebl.BedCols.ENSEMBL_ID]
                r_by_tx_by_gene[gene][tx].append(r.fields)

        with file_transaction(work_dir, self.wgs_bed_fpath) as tx:
            with open(tx, 'w') as out:
                for gname, r_by_tx in r_by_tx_by_gene.iteritems():
                    all_tx = (x for xx in r_by_tx.itervalues() for x in xx if x[ebl.BedCols.FEATURE] == 'transcript')
                    tx_sorted_list = [x[ebl.BedCols.ENSEMBL_ID] for x in sorted(all_tx, key=tx_priority_sort_key)]
                    if not tx_sorted_list:
                        continue
                    tx_id = tx_sorted_list[0]
                    for r in sorted(r_by_tx[tx_id], key=get_sort_key(chr_order)):
                        out.write('\t'.join(str(f) for f in r) + '\n')
        return self.wgs_bed_fpath


#!/usr/bin/env python

import os
import shutil
from optparse import OptionParser, SUPPRESS_HELP
from collections import defaultdict, OrderedDict
from os.path import isfile, join, basename
from tempfile import mkdtemp

from ngs_utils.logger import debug
from ngs_utils.bed_utils import verify_bed, SortableByChrom, count_bed_cols, sort_bed, clean_bed
from ngs_utils.file_utils import file_transaction, adjust_path, safe_mkdir, verify_file
from ngs_utils.logger import critical, info
from ngs_utils import logger

import ensembl as ebl
from bed_annotation.bed_annotation import annotate


# TODO: prefer consecutive annotations


def main():
    options = [
        (['-o', '--output-file'], dict(
            dest='output_file',
            metavar='FILE',
            help='Output file',
        )),
        (['--output-features'], dict(
            dest='output_features',
            action='store_true',
            default=False,
            help='Also output featues that was used to annotate',
        )),
        (['--reuse'], dict(
            dest='reuse_intermediate',
            action='store_true',
            help='reuse intermediate non-empty files in the work dir from previous run',
        )),
        (['-g', '--genome'], dict(
            dest='genome',
            help='Genome build. Accepted values: ' + ', '.join(ebl.SUPPORTED_GENOMES),
        )),
        (['--canonical'], dict(
            dest='only_canonical',
            action='store_true',
            default=False,
            help='Use only features from canonical transcripts to annotate',
        )),
        (['--short'], dict(
            dest='short',
            action='store_true',
            default=False,
            help='Add only "Gene" column (outputa 4-col BED file instead of 6-col)',
        )),
        (['--cds-only'], dict(
            dest='cds_only',
            action='store_true',
            default=False,
            help='Use only CDS to annotate',
        )),
        (['-e', '--extended'], dict(
            dest='extended',
            action='store_true',
            default=False,
            help='Output additional columns: transcript, GC, overlap size...',
        )),
        (['--high-confidence'], dict(
            dest='high_confidence',
            action='store_true',
            default=False,
            help='Annotate with only high confidence regions (TSL is 1 or NA, with HUGO symbol, total overlap size > 50%)',
        )),
        (['-a', '--ambiguities'], dict(
            dest='ambiguities',
            choices=['best_one', 'best_ask', 'best_all', 'all_ask', 'all'],
            default='best_all',
            help='How to resolve ambuguios overlaps with reliable transcripts for a single region',
        )),
        (['--seq2c'], dict(
            dest='seq2c',
            action='store_true',
            default=False,
            help=SUPPRESS_HELP,  # For back-compability
        )),
        (['-d', '--debug'], dict(
            dest='debug',
            action='store_true',
            default=False,
            help=SUPPRESS_HELP,
         )),
        (['--not-collapse-exons'], dict(
            dest='collapse_exons',
            action='store_false',
            default=True,
            help=SUPPRESS_HELP,
         )),
        (['--work-dir'], dict(dest='work_dir', metavar='DIR', help=SUPPRESS_HELP)),
        (['--log-dir'], dict(dest='log_dir', metavar='DIR', help=SUPPRESS_HELP)),
    ]

    parser = OptionParser(description='Annotating BED file based on reference features annotations.')
    for args, kwargs in options:
        parser.add_option(*args, **kwargs)
    opts, args = parser.parse_args()
    logger.init(opts.debug)

    if not opts.genome:
        critical('Error: please, specify genome build name with -g (e.g. `-g hg19`)')

    output_features             = opts.output_features
    cds_only                    = opts.cds_only
    collapse_exons              = opts.collapse_exons
    extended                    = opts.extended
    short                       = opts.short
    high_confidence             = opts.high_confidence
    only_canonical              = opts.only_canonical
    ambiguities_method          = opts.ambiguities
    if short:
        if extended:        critical('--short and --extended can\'t be set both')
        if output_features: critical('--short and --output-features can\'t be set both')
    elif output_features or extended:
        extended = True
        short = False

    if len(args) < 1:
        parser.exit('Usage: ' + __file__ + ' Input_BED_file -g hg19 -o Annotated_BED_file [options]')
    input_bed_fpath = verify_file(args[0], is_critical=True, description='Input BED file for ' + __file__)
    output_fpath = adjust_path(opts.output_file)

    if opts.work_dir:
        work_dir = join(adjust_path(opts.work_dir), os.path.splitext(basename(input_bed_fpath))[0])
        safe_mkdir(work_dir)
        info('Created work directory ' + work_dir)
    else:
        work_dir = mkdtemp('bed_annotate')
        debug('Created temporary work directory ' + work_dir)

    input_bed_fpath = clean_bed(input_bed_fpath, work_dir)
    input_bed_fpath = verify_bed(input_bed_fpath, is_critical=True, description='Input BED file for ' + __file__)

    output_fpath = annotate(
        input_bed_fpath, output_fpath, work_dir, genome=opts.genome,
        only_canonical=only_canonical, short=short, extended=extended,
        high_confidence=high_confidence, collapse_exons=collapse_exons,
        output_features=output_features, cds_only=cds_only,
        ambiguities_method=ambiguities_method)

    if not opts.work_dir:
        debug('Removing work directory ' + work_dir)
        shutil.rmtree(work_dir)

    info('Done, saved to ' + output_fpath)


if __name__ == '__main__':
    main()
